'''Implementation of atarivault: Atari Vault'''
import logging
import os
from gex.lib.tasks.basetask import BaseTask
from gex.lib.tasks import helpers
from gex.lib.tasks.impl.atarivault import arcade

logger = logging.getLogger('gextoolbox')

class AtariVaultTask(BaseTask):
    '''Implements atarivault: Atari Vault'''
    _task_name = "atarivault"
    _title = "Atari Vault"
    _details_markdown = '''
ROM files are in the installation folder, but arcade ROMS need rebuilt and only work in MAME 2003.
Some arcade ROMs based on https://gist.githubusercontent.com/cxx/6d1d44ce4a6107ed80e0a6c8c5b887c4/raw/d3cae583024f44c06d69a55d867066084838c2f7/avault2mame.js
    '''
    _default_input_folder = helpers.gen_steam_app_default_folder("Atari Vault")
    _input_folder_desc = "Atari Vault install folder"
    _prop_info = {
        "include-2600": {
            "description": "Include Atari 2600 Games",
            "default": True,
            "type": "Boolean"
        },
        "include-5200": {
            "description": "Include Atari 5200 Games",
            "default": True,
            "type": "Boolean"
        },
        "include-mnetwork": {
            "description": "Include Atari MNetwork Games",
            "default": True,
            "type": "Boolean"
        },
        "include-prototype": {
            "description": "Include Atari Prototype Games",
            "default": True,
            "type": "Boolean"
        },
        "include-arcade": {
            "description": "Include Atari Arcade Games",
            "default": True,
            "type": "Boolean"
        },
        "include-arcade-partials": {
            "description": "Include Atari Arcade Games that cannot be converted to MAME",
            "default": False,
            "type": "Boolean"
        }
    }

    def get_out_file_info(self):
        '''Return a list of output files'''
        return {
            "files": self._metadata['out']['files'],
            "notes": self._metadata['out']['notes']
        }

    def execute(self, in_dir, out_dir):
        if self._props.get('include-2600'):
            self.extract_console_set(in_dir, out_dir, '2600')
        if self._props.get('include-5200'):
            self.extract_console_set(in_dir, out_dir, '5200')
        if self._props.get('include-prototype'):
            self.extract_console_set(in_dir, out_dir, 'prototype')
        if self._props.get('include-mnetwork'):
            self.extract_console_set(in_dir, out_dir, 'mnetwork')
        if self._props.get('include-arcade'):
            include_partials = self._props.get('include-arcade-partials')
            arcade.extract(in_dir, out_dir, self, include_partials)

        logger.info("Processing complete.")

    
    def extract_console_set(self, in_dir, out_dir, game_set):
        '''Extract Atari Console ROMs'''
            
        for file_metadata in self._metadata['in']['files'].values():
            resolved_file = self.read_datafile(in_dir, file_metadata)
            if 'copy_to' in file_metadata:
                out_file_entries = [x for x in self._metadata['out']['files'] if x['game'] == file_metadata['copy_to'] and x['set'] == game_set]
                # Set isn't used in verification. Make 2600/5200/Arcade rom names unique, but keep the set match here
                if len(out_file_entries) == 1:
                    out_file_entry = out_file_entries[0]
                    logger.info(f"Copying {out_file_entry['game']}...")
                    filename = out_file_entry['filename']
                    _ = self.verify_out_file(filename, resolved_file['contents'])
                    out_path = os.path.join(out_dir, filename)
                    with open(out_path, "wb") as out_file:
                        out_file.write(resolved_file['contents'])




'''Implementation of ddragontrilogy: Double Dragon Trilogy'''
import glob
import logging
import os
from gex.lib.utils.blob import transforms
from gex.lib.tasks import helpers
from gex.lib.tasks.basetask import BaseTask
from gex.lib.utils import gfx_rebuilder

logger = logging.getLogger('gextoolbox')

class DoubleDragonTrilogyTask(BaseTask):
    '''Implements ddragontrilogy: Double Dragon Trilogy'''
    _task_name = "ddragontrilogy"
    _title = "Double Dragon Trilogy"
    _details_markdown = '''
Based on dotemu2mame.js: https://gist.github.com/cxx/81b9f45eb5b3cb87b4f3783ccdf8894f
'''
    _default_input_folder = helpers.gen_steam_app_default_folder("Double Dragon Trilogy")
    _input_folder_desc = "Double Dragon Trilogy Steam folder"

    def get_out_file_info(self):
        '''Return a list of output files'''
        return {
            "files": self._metadata['out']['files'],
            "notes": self._metadata['out']['notes']
        }

    def execute(self, in_dir, out_dir):
        verified_in_files = self.read_all_datafiles(in_dir)
        in_files = {k: v['contents'] for k,v in verified_in_files.items()}

        funcs = [
            self._ddragon,
            self._ddragon2,
            self._ddragon3
        ]

        for func in funcs:
            logger.info(f"Extracting {func.__name__[1:]}...")
            rom_package = func(in_files)
            _ = self.verify_out_file(rom_package['filename'], rom_package['contents'])
            with open(os.path.join(out_dir, rom_package['filename']), "wb") as out_file:
                out_file.write(rom_package['contents'])

        logger.info("Processing complete.")

    def _dotemu_reencode_gfx_helper(self, in_file_name, filenames, layout):
        def encode(in_files):
            contents = in_files[in_file_name]
            contents = gfx_rebuilder.reencode_gfx(contents, layout)
            chunks = transforms.equal_split(contents, num_chunks=len(filenames))
            return dict(zip(filenames, chunks))
        return encode

    _DDRAGON_CHAR_LAYOUT = {
        'width': 8,
        'height': 8,
        'total': [1,1],
        'planes': 4,
        'planeoffset': [0, 2, 4, 6],
        'xoffset': [1, 0, 8*8+1, 8*8+0, 16*8+1, 16*8+0, 24*8+1, 24*8+0],
        'yoffset': [0*8, 1*8, 2*8, 3*8, 4*8, 5*8, 6*8, 7*8],
        'charincrement': 32*8
    }

    _DDRAGON_TILE_LAYOUT = {
        'width': 16,
        'height': 16,
        'total': [1,2],
        'planes': 4,
        'planeoffset': [[1,2,0], [1,2,4], 0, 4],
        'xoffset': [3, 2, 1, 0, 16*8+3, 16*8+2, 16*8+1, 16*8+0,
            32*8+3, 32*8+2, 32*8+1, 32*8+0, 48*8+3, 48*8+2, 48*8+1, 48*8+0],
        'yoffset': [0*8, 1*8, 2*8, 3*8, 4*8, 5*8, 6*8, 7*8,
            8*8, 9*8, 10*8, 11*8, 12*8, 13*8, 14*8, 15*8],
        'charincrement': 64*8
    }

    def _ddragon(self, in_files):
        func_map = {}

        # MainCPU
        maincpu_filenames = [
            '21j-1-5.26',
            '21j-2-3.25',
            '21j-3.24',
            '21j-4-1.23'
        ]
        func_map['maincpu'] = helpers.equal_split_helper(
            'ddragon_hd6309.bin',
            maincpu_filenames
        )

        # Sub
        func_map['sub'] = helpers.name_file_helper(
            'ddragon_hd63701.bin',
            '21jm-0.ic55'
        )

        # SoundCPU
        func_map['soundcpu'] = helpers.name_file_helper(
            'ddragon_m6809.bin',
            '21j-0-1'
        )

        # Gfx 1
        gfx1_filesnames = [
            '21j-5'
        ]
        func_map['gfx1'] = self._dotemu_reencode_gfx_helper(
            'ddragon_gfxdata1.bin',
            gfx1_filesnames,
            self._DDRAGON_CHAR_LAYOUT)

        # Gfx 2
        gfx2_filesnames = [
            '21j-a',
            '21j-b',
            '21j-c',
            '21j-d',
            '21j-e',
            '21j-f',
            '21j-g',
            '21j-h'
        ]
        func_map['gfx2'] = self._dotemu_reencode_gfx_helper(
            'ddragon_gfxdata2.bin',
            gfx2_filesnames,
            self._DDRAGON_TILE_LAYOUT
        )

        # Gfx 3
        gfx3_filesnames = [
            '21j-8',
            '21j-9',
            '21j-i',
            '21j-j'
        ]
        func_map['gfx3'] = self._dotemu_reencode_gfx_helper(
            'ddragon_gfxdata3.bin',
            gfx3_filesnames,
            self._DDRAGON_TILE_LAYOUT
        )

        # ADPCM
        adpcm_filenames = [
            '21j-6',
            '21j-7'
        ]
        func_map['adpcm'] = helpers.equal_split_helper('ddragon_adpcm.bin', adpcm_filenames)

        # PROMs
        prom_file_map = {
            '21j-k-0': 0x100,
            '21j-l-0': 0x200
        }
        func_map['prom'] = helpers.custom_split_helper('proms.bin', prom_file_map)

        return {'filename': 'ddragon.zip', 'contents': helpers.build_rom(in_files, func_map)}

    def _ddragon2(self, in_files):
        func_map = {}

        # MainCPU
        maincpu_filenames = [
            '26a9-04.bin',
            '26aa-03.bin',
            '26ab-0.bin',
            '26ac-0e.63'
        ]
        func_map['maincpu'] = helpers.equal_split_helper('ddragon2_hd6309.bin', maincpu_filenames)

        # Sub
        func_map['sub'] = helpers.name_file_helper(
            'ddragon2_z80sub.bin',
            '26ae-0.bin'
        )

        # SoundCPU
        func_map['soundcpu'] = helpers.name_file_helper(
            'ddragon2_z80sound.bin',
            '26ad-0.bin'
        )

        # Gfx 1
        gfx1_filesnames = [
            '26a8-0e.19'
        ]
        func_map['gfx1'] = self._dotemu_reencode_gfx_helper(
            'ddragon2_gfxdata1.bin',
            gfx1_filesnames,
            self._DDRAGON_CHAR_LAYOUT
        )

        # Gfx 2
        gfx2_filesnames = [
            '26j0-0.bin',
            '26j1-0.bin',
            '26af-0.bin',
            '26j2-0.bin',
            '26j3-0.bin',
            '26a10-0.bin'
        ]
        func_map['gfx2'] = self._dotemu_reencode_gfx_helper(
            'ddragon2_gfxdata2.bin',
            gfx2_filesnames,
            self._DDRAGON_TILE_LAYOUT
        )

        # Gfx 3
        gfx3_filesnames = [
            '26j4-0.bin',
            '26j5-0.bin'
        ]
        func_map['gfx3'] = self._dotemu_reencode_gfx_helper(
            'ddragon2_gfxdata3.bin',
            gfx3_filesnames,
            self._DDRAGON_TILE_LAYOUT
        )

        # OKI
        oki_filenames = [
            '26j6-0.bin',
            '26j7-0.bin'
        ]
        func_map['oki'] = helpers.equal_split_helper(
            'ddragon2_oki.bin',
            oki_filenames
        )

        # PROMs
        prom_file_map = {
            '21j-k-0': 0x100,
            'prom.16': 0x200
        }
        func_map['prom'] = helpers.custom_split_helper(
            'proms.bin',
            prom_file_map
        )

        return {'filename': 'ddragon2.zip', 'contents': helpers.build_rom(in_files, func_map)}


    _WWF_TILE_LAYOUT = {
        'width': 16,
        'height': 16,
        'total': [1,2],
        'planes': 4,
        'planeoffset': [8, 0, [1,2,8], [1,2,0]],
        'xoffset': [0, 1, 2, 3, 4, 5, 6, 7,
                  32*8+0, 32*8+1, 32*8+2, 32*8+3, 32*8+4, 32*8+5, 32*8+6, 32*8+7],
        'yoffset': [0*16, 1*16, 2*16, 3*16, 4*16, 5*16, 6*16, 7*16,
    	      16*8, 16*9, 16*10, 16*11, 16*12, 16*13, 16*14, 16*15],
        'charincrement': 64*8
    }

    _WWF_SPRITE_LAYOUT = {
        'width': 16,
        'height': 16,
        'total': [1,4],
        'planes': 4,
        'planeoffset': [[0,4], [1,4], [2,4], [3,4]],
        'xoffset': [0, 1, 2, 3, 4, 5, 6, 7,
                  16*8+0, 16*8+1, 16*8+2, 16*8+3, 16*8+4, 16*8+5, 16*8+6, 16*8+7],
        'yoffset': [0*8, 1*8, 2*8, 3*8, 4*8, 5*8, 6*8, 7*8,
    	      8*8, 9*8, 10*8, 11*8, 12*8, 13*8, 14*8, 15*8],
        'charincrement': 32*8
    }

    def _ddragon3(self, in_files):
        func_map = {}

        # CPU
        maincpu_filenames = [
            '30a15-0.ic79',
            '30a14-0.ic78'
        ]
        def maincpu(in_files):
            contents = in_files['ddragon3_m68k.bin']
            chunks = transforms.deinterleave(contents, num_ways=2, word_size=1)
            chunks[0] = chunks[0][0:0x20000]
            return dict(zip(maincpu_filenames, chunks))
        func_map['maincpu'] = maincpu

        # AudioCPU
        func_map['audiocpu'] = helpers.name_file_helper('ddragon3_z80.bin', '30a13-0.ic43')

        # Gfx 1
        gfx1_filenames = [
            '30j-6.ic5',
            '30j-4.ic7',
            '30j-7.ic4',
            '30j-5.ic6'
        ]
        def gfx1(in_files):
            contents = in_files['ddragon3_gfxdata1.bin']
            contents = gfx_rebuilder.reencode_gfx(contents, self._WWF_TILE_LAYOUT)
            chunks = transforms.deinterleave(contents, num_ways=2, word_size=1)
            contents = transforms.merge(chunks)
            chunks = transforms.equal_split(contents, len(gfx1_filenames))
            return dict(zip(gfx1_filenames, chunks))
        func_map['gfx1'] = gfx1

        # Gfx 2
        gfx2_filenames = [
            '30j-3.ic9',
            '30a12-0.ic8',
            '30j-2.ic11',
            '30a11-0.ic10',
            '30j-1.ic13',
            '30a10-0.ic12',
            '30j-0.ic15',
            '30a9-0.ic14'
        ]
        def gfx2(in_files):
            contents = in_files['ddragon3_gfxdata2.bin']
            contents = gfx_rebuilder.reencode_gfx(contents, self._WWF_SPRITE_LAYOUT)
            chunks = transforms.custom_split(contents,
                [0x80000, 0x10000, 0x80000, 0x10000, 0x80000, 0x10000, 0x80000, 0x10000])
            return dict(zip(gfx2_filenames, chunks))
        func_map['gfx2'] = gfx2

        # OKI
        func_map['oki'] = helpers.name_file_helper('ddragon3_oki.bin', '30j-8.ic73')

        # PROMs
        def dummy_prom(_):
            return {'mb7114h.ic38': bytearray(0x100)}
        func_map['prom'] = dummy_prom

        return {'filename': 'ddragon3.zip', 'contents': helpers.build_rom(in_files, func_map)}



import os
import logging
import click
import click_log

from gex.lib.utils import helper

logger = logging.getLogger('gextoolbox')

@click.command()
@click.option('--srcdir', 'src_dir',
    help = 'path required by the transform set - see task for details', default=None)
@click.option('--destdir', 'dest_dir', help = 'path to send reassembled ROMs to', required=True)
@click.option('--task', 'task', help = 'name of the transform set to run', required=True)
@click.option("--prop", 'props', multiple=True, help = 'per task config options; see task details for supported options', default=[])
@click_log.simple_verbosity_option(logger)
def extract(src_dir, dest_dir, task, props):
    """Run a task to extract roms from Steam/GOG/etc. games"""

    # Load the task module
    task_class = helper.load_task(task)

    # If there isn't a src_dir set, pull in the default
    if not src_dir:
        src_dir = task_class.get_default_input_folder()
        # If there isn't a default, exit
        if not src_dir:
            logger.error(f"Task {task} requires a source dir; see task details for more info")
            exit()

    # Make sure input dir exists
    if not os.path.exists(src_dir):
        logger.error(f"Source dir {src_dir} does not exist; see task details for more info")
        exit()

    # Ensure the output folder exists or can be made
    helper.preparepath(dest_dir)

    task_class.read_task_metadata()

    task_class.set_props(props)

    task_class.execute(src_dir, dest_dir)



'''Implementation of blizzarcade: Blizzard Arcade Collection'''
from gex.lib.tasks.copytask import CopyTask
from gex.lib.tasks import helpers

class BubsyTask(CopyTask):
    '''Implements blizzarcade: Blizzard Arcade Collection'''
    _task_name = "blizzarcade"
    _title = "Blizzard Arcade Collection"
    _details_markdown = '''
These are the ROMs just sitting in the install folder. 
    '''
    _default_input_folder = r"C:\Program Files (x86)\Blizzard Arcade Collection"
    _input_folder_desc = "Blizzard Arcade Collection install folder"



'''Implementation of disneyalkb: Disney Aladdin / Lion King Bundle (and DLC)'''
import glob
import logging
import os
from gex.lib.contrib.bputil import BPListReader
from gex.lib.tasks import helpers
from gex.lib.tasks.basetask import BaseTask

logger = logging.getLogger('gextoolbox')

class DisneyClassicsTask(BaseTask):
    '''Implements disneyalkb: Disney Aladdin / Lion King Bundle (and DLC)'''
    _task_name = "disneyalkb"
    _title = "Disney Aladdin / Lion King Bundle (and DLC)"
    _details_markdown = '''
Based on https://github.com/farmerbb/RED-Project/wiki/Disney-Classic-Games:-Aladdin-and-The-Lion-King
'''
    _default_input_folder = helpers.gen_steam_app_default_folder(
        "Disney Classic Games Aladdin and the Lion King")
    _input_folder_desc = "Disney Classics Steam folder"

    def get_out_file_info(self):
        '''Return a list of output files'''
        return {
            "files": self._metadata['out']['files'],
            "notes": self._metadata['out']['notes']
        }

    def execute(self, in_dir, out_dir):
        for file_metadata in self._metadata['in']['files'].values():
            file_name = file_metadata['filename']
            pkg_name = self._pkg_name_map.get(file_name)
            datafile = self.read_datafile(in_dir, file_metadata)
            contents = datafile['contents']
            logger.info(f'Reading files for {file_name}...')
            reader = BPListReader(contents)
            parsed = reader.parse()

            handler_func = self.find_handler_func(pkg_name)
            if parsed is not None and handler_func is not None:
                output_files = handler_func(parsed)
                for out_file_entry in output_files:
                    _ = self.verify_out_file(out_file_entry['filename'], out_file_entry['contents'])
                    out_path = os.path.join(out_dir, out_file_entry['filename'])
                    with open(out_path, "wb") as out_file:
                        out_file.write(out_file_entry['contents'])
            elif parsed is None:
                logger.warning("Could not find merged rom data in mbundle.")
            elif handler_func is None:
                logger.warning("Could not find matching handler function.")
        logger.info("Processing complete.")

    _pkg_name_map = {
        'bundleAladdin.mbundle': 'aladdin',
        'bundleDLC1.mbundle': 'dlc',
        'bundleLionKing.mbundle': 'lionking',
        'bundleMain.mbundle': 'main',
    }

    def _handle_aladdin(self, mbundle_entries):
        files = {
            'Aladdin-Remix.bin',
            'Aladdin.4F7A.PATCHED.bin',
            'Aladdin.9CB2.PATCHED.bin',
            'Aladdin.CES.PATCHED.bin',
            'Aladdin.gb'
        }
        return self._bundle_handler(files, mbundle_entries)

    def _handle_dlc(self, mbundle_entries):
        files = {
            'JungleBook.gb',
            'JungleBook.md',
            'JungleBook.nes',
            'Aladdin.sfc',
            'JungleBook.sfc',
            'JungleBookMusicROM.sfc'
        }
        return self._bundle_handler(files, mbundle_entries)

    def _handle_lionking(self, mbundle_entries):
        files = {
            'LionKing.0D3D.PATCHED.bin',
            'LionKing.gb',
            'LionKing.919A.PATCHED.sfc',
            'LionKing.DE6E.PATCHED.sfc',
        }
        return self._bundle_handler(files, mbundle_entries)

    def _handle_main(self, mbundle_entries):
        files = {
            'JUNGLEBOOK.PATCHED.md',
        }
        return self._bundle_handler(files, mbundle_entries)

    def _bundle_handler(self, files, mbundle_entries):
        out_files = []
        for file in files:
            out_name = file.replace('.bin', '.md')
            logger.info(f'Extracting {out_name}...')
            out_files.append({'filename': out_name, 'contents': mbundle_entries[file]})
        return out_files
        


'''
Utility functions for SNK games
'''

# from gex.lib.utils.blob import transforms

def sfix_reorder(contents):
    '''Reorders an SFIX rom file'''
    for i in range(0, len(contents), 32):
        tmp = bytearray(32)
        for j in range(0, 8):
            tmp[j+16] = contents[i+4*j+0]
            tmp[j+24] = contents[i+4*j+1]
            tmp[j+ 0] = contents[i+4*j+2]
            tmp[j+ 8] = contents[i+4*j+3]
        contents[i:i+32] = tmp
    return contents



'''Contains more generic utility functions'''

import importlib
import inspect
import os

from gex.lib.tasks.basetask import BaseTask

def cleanpath(path_str):
    '''Clean the double-quotes from a path'''
    if path_str.startswith("\""):
        path_str = path_str[1:]
    if path_str.endswith("\""):
        path_str = path_str[:-1]
    return path_str

def preparepath(out_path):
    '''Make sure the output folder exists and is writable or can be created'''
    if not os.path.exists(out_path):
        try:
            os.makedirs(out_path)
        except OSError as error:
            raise Exception("Cannot create output folder.") from error
    else:
        if not os.access(out_path, os.W_OK):
            raise Exception("Cannot write to output folder.")

def load_task(task):
    '''Load a task module's task class from the task name'''
    package = f'gex.lib.tasks.impl.{task}'
    transform_module = importlib.import_module(package)
    clsmembers = inspect.getmembers(transform_module, inspect.isclass)
    exclude_list = [
        'BaseTask',
        'CopyTask',
        'ZipSpliceTask',
        'SpliceTask'
    ]
    for name, typedef in clsmembers:
        if name not in exclude_list and issubclass(typedef, BaseTask):
            return typedef()
    return None

def get_task_names():
    pass



import click
from rich.console import Console
from rich.markdown import Markdown
from texttable import Texttable
from gex.lib.utils import helper

@click.command()
@click.option('--task', 'task', help = 'name of the transform set to run', required=True)
def details(task):
    """Get details about a specific extraction task"""

    task_class = helper.load_task(task)
    if task_class is None:
        raise Exception(f"Task {task} not found to describe.")
    task_class.read_task_metadata()
    markdown_text = get_header_display(task_class)
    markdown_text += "\n\n"
    markdown_text += task_class.get_details_markdown()
    markdown_text += "\n\n"
    console = Console()
    formatted_text = Markdown(markdown_text)
    console.print(formatted_text)
    print(get_output_display(task_class))
    prop_table = get_prop_display(task_class)
    if prop_table:
        print(prop_table)


def get_header_display(task_class):
    '''Get a markdown-formatted header for this task for help, etc.'''
    markdown_text = f'# {task_class.get_task_name()}: {task_class.get_title()}\n'
    markdown_text += f'  Expected input dir: {task_class.get_input_folder_description()}\n\n'
    markdown_text += f'  Default input dir: "{task_class.get_default_input_folder()}"\n'
    return markdown_text

def get_output_display(task_class):
    '''Get a display table for the output files for this task'''
    out_file_info = task_class.get_out_file_info()
    table = Texttable()
    table.add_row(["Game", "System", "Status", "Filename", "Notes"])
    for item in out_file_info['files']:
        notes = " ".join([str(note) for note in item['notes']])
        table.add_row([item['game'], item['system'], item['status'], item['filename'], notes])
    output = table.draw()

    if out_file_info['notes']:
        output += "\n\n"
        for key, value in out_file_info['notes'].items():
            output += f"{key}: {value}\n"

    return output

def get_prop_display(task_class):
    '''Get a display table for the props available for this task'''
    if not task_class.get_prop_info():
        return None

    table = Texttable()
    table.add_row(["Name", "Type", "Default", "Description"])
    for key, value in task_class.get_prop_info().items():
        table.add_row([key, value['type'], str(value['default']), value['description']])
    return table.draw()



import click

from .arc.arc import arc
from .kpka.kpka import kpka
from .zip.zip import zip_cli

@click.group()
def archive():
    """Tools for working with various types of archives"""

archive.add_command(arc)
archive.add_command(kpka)
archive.add_command(zip_cli)



import os
import logging
import click
import click_log
from texttable import Texttable
from gex.lib.archive import zip as zip_lib

logger = logging.getLogger('gextoolbox')

@click.command(name='list')
@click.option('--in', 'in_file', help = 'path to input ZIP archive', required=True)
@click.option('--format', type=click.Choice(['table', 'verify-json'], case_sensitive=False), default="table")
@click_log.simple_verbosity_option(logger)
def list_files(in_file, format):
    """List the contents of a ZIP archive"""
    try:
        with open(in_file, "rb") as curr_file:
            file_content = bytearray(curr_file.read())
            zip_metas = zip_lib.get_metadata(file_content)

            format = format.lower()
            if format == "table":
                table = Texttable()
                table.add_row(["Filename", "Size", "CRC"])
                table.set_cols_dtype(["t", "t", "t"])
                for zip_meta in zip_metas.values():
                    table.add_row([zip_meta['filename'], zip_meta['size'], zip_meta['crc']])
                print(table.draw())
            elif format == "verify-json":
                print('"verify": {')
                print('    "type": "zip",')
                print('    "entries": {')
                values = list(zip_metas.values())
                for zip_meta in values:
                    line = '        "'
                    line += zip_meta.get("filename")
                    line += '": {"size": '
                    line += str(zip_meta.get("size"))
                    line += ', "crc": "'
                    line += zip_meta.get("crc")
                    line += '"}'
                    if zip_meta != values[-1]:
                        line += ','
                    print(line)
                print('    }')
                print('}')

    except Exception as error:
        logger.error(error)
        logger.error('Error While Opening the file!')



import logging
import click
import gex.lib.tasks.impl as impl
from gex.lib.utils import helper

logger = logging.getLogger('gextoolbox')

@click.command(name='list')
def list_cli():
    """List available extraction tasks"""
    for task in impl.__all__:
        task_class = helper.load_task(task)
        if task_class:
            name = task_class.get_task_name()
            if name is not None:
                logger.info(f"{name.rjust(15, ' ')}: {task_class.get_title()}")



'''Extraction code for arcade ROMS from DLC bundle'''
import logging
from gex.lib.tasks import helpers
from gex.lib.tasks.impl.snk40 import utils
from gex.lib.utils.blob import transforms

logger = logging.getLogger('gextoolbox')

def extract(bundle_contents):
    '''Extract files from DLC bundle'''
    out_files = []
    contents = bundle_contents['dlc']
    out_files.extend(_handle_bbusters(contents))
    out_files.extend(_handle_searchar(contents))

    return out_files

def _gfx_split_swap(in_file_ref, filenames):
    '''Func map helper for transforms.equal_split'''
    def split(in_files):
        contents = in_files[in_file_ref]
        chunks = transforms.equal_split(contents, num_chunks = len(filenames))
        chunks = transforms.swap_endian_all(chunks)
        return dict(zip(filenames, chunks))
    return split

def _handle_bbusters(bundle_contents):
    # bbusters commons
    func_map = {}
    out_files = []
    func_map['gfx1'] = helpers.name_file_helper('bbusters.gfx1', 'bb-10.l9')
    gfx2_filenames = [
        "bb-f11.m16",
        "bb-f12.m13",
        "bb-f13.m12",
        "bb-f14.m11"
    ]
    func_map['gfx2'] = _gfx_split_swap('bbusters.gfx2', gfx2_filenames)
    gfx3_filenames = [
        "bb-f21.l10",
        "bb-f22.l12",
        "bb-f23.l13",
        "bb-f24.l15"
    ]
    func_map['gfx3'] = _gfx_split_swap('bbusters.gfx3', gfx3_filenames)
    func_map['gfx4'] = helpers.name_file_helper('bbusters.gfx4', 'bb-back1.m4')
    func_map['gfx5'] = helpers.name_file_helper('bbusters.gfx5', 'bb-back2.m6')
    func_map['scale_table1'] = helpers.name_file_helper('bbusters.scale_table', 'bb-6.e7')
    func_map['scale_table2'] = helpers.name_file_helper('bbusters.scale_table', 'bb-7.h7')
    func_map['scale_table3'] = helpers.name_file_helper('bbusters.scale_table', 'bb-8.a14')
    func_map['scale_table4'] = helpers.name_file_helper('bbusters.scale_table', 'bb-9.c14')
    func_map['audiocpu'] = helpers.name_file_helper('bbusters.audiocpu', 'bb-1.e6')
    func_map['ymsnd'] = helpers.name_file_helper('bbusters.ymsnd', 'bb-pcma.l5')
    func_map['eeprom'] = helpers.name_file_helper('bbusters.eeprom', 'bbusters-eeprom.bin')

    logger.info("Processing bbusters common files...")
    common_file_map = helpers.process_rom_files(bundle_contents, func_map)

    # bbusters
    func_map = {}
    maincpu_filenames = [
        "bb-3.k10",
        "bb-5.k12",
        "bb-2.k8",
        "bb-4.k11"
    ]
    def bbusters_maincpu(in_file_name, filenames):
        def maincpu(in_files):
            contents = in_files[in_file_name]
            chunks = transforms.equal_split(contents, num_chunks = 2)

            new_chunks = []
            for oldchunk in chunks:
                new_chunks.extend(transforms.deinterleave(oldchunk, num_ways=2, word_size=1))
            chunks = new_chunks

            return dict(zip(filenames, chunks))
        return maincpu
    func_map['maincpu'] = bbusters_maincpu('bbusters.maincpu', maincpu_filenames)
    func_map['ymsnd.deltat'] = helpers.name_file_helper('bbusters.ymsnd.deltat', 'bb-pcmb.l3')
    func_map['common'] = helpers.existing_files_helper(common_file_map)
    out_files.append(utils.build_snk_rom("bbusters.zip", bundle_contents, func_map))

    # bbustersj
    func_map = {}
    maincpu_filenames = [
        "bb3_ver2_j3.k10",
        "bb5_ver2_j3.k12",
        "bb-2.k8",
        "bb-4.k11"
    ]
    func_map['maincpu'] = bbusters_maincpu('bbustersj.maincpu', maincpu_filenames)
    func_map['ymsnd.deltat'] = helpers.name_file_helper('bbusters.ymsnd.deltat', 'bb-pcmb.l3')
    func_map['common'] = helpers.existing_files_helper(common_file_map)
    out_files.append(utils.build_snk_rom("bbustersj.zip", bundle_contents, func_map))

    # bbustersu
    func_map = {}
    maincpu_filenames = [
        "bb-ver3-u3.k10",
        "bb-ver3-u5.k12",
        "bb-2.k8",
        "bb-4.k11"
    ]
    func_map['maincpu'] = bbusters_maincpu('bbustersu.maincpu', maincpu_filenames)
    func_map['common'] = helpers.existing_files_helper(common_file_map)
    out_files.append(utils.build_snk_rom("bbustersu.zip", bundle_contents, func_map))
    return out_files

def _handle_searchar(bundle_contents):
    # searchar common
    func_map = {}
    out_files = []
    func_map['soundcpu'] = helpers.name_file_helper('searchar.soundcpu', 'bh.5')
    gfx1_filenames = [
        "bh.7",
        "bh.8"
    ]
    func_map['gfx1'] = helpers.equal_split_helper("searchar.gfx1", gfx1_filenames)
    gfx2_filenames = [
        "bh.c1",
        "bh.c3",
        "bh.c5",
        "bh.c2",
        "bh.c4",
        "bh.c6"
    ]
    def searchar_gfx2(in_file_name, filenames):
        def gfx2(in_files):
            contents = in_files[in_file_name]
            chunks = transforms.equal_split(contents, 7)
            del chunks[3] # Middle element is junk
            return dict(zip(filenames, chunks))
        return gfx2
    func_map['gfx2'] = searchar_gfx2("searchar.gfx2", gfx2_filenames)
    func_map['upd'] = helpers.name_file_helper('searchar.upd', 'bh.v1')

    logger.info("Processing searchar common files...")
    common_file_map = helpers.process_rom_files(bundle_contents, func_map)

    # searchar
    func_map = {}
    maincpu_filenames = [
        "bhw.2",
        "bhw.3"
    ]
    def searchar_maincpu(in_file_name, filenames):
        def maincpu(in_files):
            contents = in_files[in_file_name]
            chunks = transforms.deinterleave(contents, num_ways=2, word_size=1)
            return dict(zip(filenames, chunks))
        return maincpu
    func_map['maincpu'] = searchar_maincpu('searchar.maincpu', maincpu_filenames)
    user_filenames = [
        "bhw.1",
        "bhw.4"
    ]
    func_map['user1'] = searchar_maincpu('searchar.user1', user_filenames)
    func_map['common'] = helpers.existing_files_helper(common_file_map)
    out_files.append(utils.build_snk_rom("searchar.zip", bundle_contents, func_map))

    # searcharj
    func_map = {}
    maincpu_filenames = [
        "bh2ver3j.9c",
        "bh3ver3j.10c"
    ]
    func_map['maincpu'] = searchar_maincpu('searcharj.maincpu', maincpu_filenames)
    user_filenames = [
        "bhw.1",
        "bhw.4"
    ]
    func_map['user1'] = searchar_maincpu('searchar.user1', user_filenames)
    func_map['common'] = helpers.existing_files_helper(common_file_map)
    out_files.append(utils.build_snk_rom("searcharj.zip", bundle_contents, func_map))

    # searcharu
    func_map = {}
    maincpu_filenames = [
        "bh.2",
        "bh.3"
    ]
    func_map['maincpu'] = searchar_maincpu('searcharu.maincpu', maincpu_filenames)
    user_filenames = [
        "bh.1",
        "bh.4"
    ]
    func_map['user1'] = searchar_maincpu('searcharu.user1', user_filenames)
    func_map['common'] = helpers.existing_files_helper(common_file_map)
    out_files.append(utils.build_snk_rom("searcharu.zip", bundle_contents, func_map))
    return out_files



'''Tools for Graphics ROMs, largely based on dotemu2mame.js and avault2mame.js by cxx on GitHub'''
import copy
from collections import namedtuple
import struct

# TODO: Make these functions easier to read/more pythonic

def reencode_gfx(contents, layout):
    '''
    Rebuild GFX roms from sprite/tile content.
    # Example Layout
    _DDRAGON_TILE_LAYOUT = {
        'width': 16,
        'height': 16,
        'total': [1,2],
        'planes': 4,
        'planeoffset': [[1,2,0], [1,2,4], 0, 4],
        'xoffset': [3, 2, 1, 0, 16*8+3, 16*8+2, 16*8+1, 16*8+0,
            32*8+3, 32*8+2, 32*8+1, 32*8+0, 48*8+3, 48*8+2, 48*8+1, 48*8+0],
        'yoffset': [0*8, 1*8, 2*8, 3*8, 4*8, 5*8, 6*8, 7*8,
            8*8, 9*8, 10*8, 11*8, 12*8, 13*8, 14*8, 15*8],
        'charincrement': 64*8
    }
'''
    mod_layout = copy.deepcopy(layout)
    num_planes = mod_layout['planes']
    dest = bytearray(len(contents) * num_planes // 8)

    if isinstance(mod_layout['total'], list):
        [num, den] = mod_layout['total']
        temp_layout = copy.deepcopy(mod_layout)
        temp_layout['total'] = len(dest) * 8 // mod_layout['charincrement'] * num // den

        def map_plane_offset(plane_offset):
            if isinstance(plane_offset, list):
                [num, den, *add] = plane_offset
                add = add[0] if add else 0
                return len(dest) * 8 * num // den + add
            else:
                return plane_offset
        temp_layout['planeoffset'] = list(map(map_plane_offset, mod_layout['planeoffset']))
        mod_layout = temp_layout

    i = 0
    for curr_char in range(0, mod_layout['total']):
        charoffset = mod_layout['charincrement'] * curr_char
        for curr_height in range(0, mod_layout['height']):
            yoffset = charoffset + mod_layout['yoffset'][curr_height]
            for curr_width in range(0, mod_layout['width']):
                xoffset = yoffset + mod_layout['xoffset'][curr_width]
                for curr_plane in range(0, num_planes):
                    offset = xoffset + mod_layout['planeoffset'][curr_plane]
                    dest[offset >> 3] = dest[offset >> 3] | ((contents[i] >> num_planes-1-curr_plane) & 1) << (~offset & 7)
                i += 1

    return dest

BitmapMeta = namedtuple('BitmapMeta', 'offset width height')
def reverse_bmp(contents):
    '''Given a .BMP bitmap, reverse it for gfx rom rebuilding'''
    bmp_meta = BitmapMeta._make(struct.unpack('<L4xLL', contents[10:26]))
    buf = bytearray()
    for row in range(0, bmp_meta.height):
        src_start = bmp_meta.offset + (bmp_meta.width * (bmp_meta.height - (1 + row)))
        buf += contents[src_start:src_start + bmp_meta.width]
    return buf


import logging
import os
import click
import click_log
from gex.lib.utils.blob import hash as hash_helper

logger = logging.getLogger('gextoolbox')

@click.command(name="hash")
@click.option('--in', 'in_file', help = 'path to input file', required=True)
@click.option('--type', 'hash_type', help = 'type of checksum/hash to calculate',
    type=click.Choice(['CRC', 'MD5', 'SHA1'], case_sensitive=False), required=True)
@click_log.simple_verbosity_option(logger)
def hash_cli(in_file, hash_type):
    '''CLI tool to hash a file with the specified algorithm'''
    try:
        with open(in_file, "rb") as file:
            in_data = file.read()
    except IOError:
        logger.error(f"Error reading {in_file}!")
        exit()

    if hash_type == 'CRC':
        checksum = hash_helper.get_crc(in_data)
    elif hash_type == 'MD5':
        checksum = hash_helper.get_md5(in_data)
    elif hash_type == 'SHA1':
        checksum = hash_helper.get_sha1(in_data)

    logger.info(f'{os.path.basename(in_file)} {hash_type}: {checksum}')



'''Extraction code for Arcade ROMs from SNK40 Patch Bundle'''
import logging
from gex.lib.tasks import helpers
from gex.lib.tasks.impl.snk40 import utils
from gex.lib.utils.blob import transforms

logger = logging.getLogger('gextoolbox')

def extract(bundle_contents):
    '''Extract Arcade ROMs from patch bundle'''
    out_files = []
    contents = bundle_contents['patch']
    out_files.extend(_handle_chopper(contents))
    out_files.extend(_handle_fantasy(contents))
    out_files.extend(_handle_timesoldiers(contents))
    out_files.extend(_handle_munchmobile(contents))
    out_files.extend(_handle_sasuke(contents))
    out_files.extend(_handle_ozmawars(contents))
    out_files.extend(_handle_paddlemania(contents))
    out_files.extend(_handle_bermuda(contents))
    out_files.extend(_handle_worldwars(contents))
    out_files.extend(_handle_marvin(contents))
    return out_files

def _handle_chopper(bundle_contents):
    out_files = []

    # CHOPPERB
    func_map = {}
    func_map['maincpu'] = helpers.name_file_helper("chopper.maincpu", "kk1.8g")
    func_map['sub'] = helpers.name_file_helper("chopper.sub", "kk4.6g")
    prom_file_names = [
        "1.9w",
        "3.9u",
        "2.9v"
    ]
    func_map['prom'] = helpers.equal_split_helper("chopper.proms", prom_file_names)
    func_map['audiocpu'] = helpers.name_file_helper("chopper.audiocpu", "kk3.3d")
    func_map['tx'] = helpers.name_file_helper("chopper.tx_tiles", "kk5.8p")
    bg_file_names = [
        "kk10.8y",
        "kk11.8z",
        "kk12.8ab",
        "kk13.8ac"
    ]
    func_map['bg'] = helpers.equal_split_helper("chopper.bg_tiles", bg_file_names)
    sp16_file_names = [
        "kk9.3k",
        "kk8.3l",
        "kk7.3n",
        "kk6.3p"
    ]
    func_map['sp16'] = helpers.equal_split_helper("chopper.sp16_tiles", sp16_file_names)
    sp32_file_names = [
        "kk18.3ab",
        "kk19.2ad",
        "kk20.3y",
        "kk21.3aa",
        "kk14.3v",
        "kk15.3x",
        "kk16.3s",
        "kk17.3t"
    ]
    func_map['sp32'] = helpers.equal_split_helper("chopper.sp32_tiles", sp32_file_names)
    func_map['ym2'] = helpers.name_file_helper("chopper.ym2", "kk2.3j")
    func_map['plds'] = helpers.name_file_helper("chopper.plds", "p-a1.2c")
    out_files.append(utils.build_snk_rom("chopperb.zip", bundle_contents, func_map))

    # LEGOFAIR
    func_map = {}
    func_map['maincpu'] = helpers.name_file_helper("legofair.maincpu", "kk1.4m")
    func_map['sub'] = helpers.name_file_helper("legofair.sub", "kk4.8m")
    prom_file_names = [
        "1.1k",
        "2.1l",
        "3.2k"
    ]
    func_map['prom'] = helpers.equal_split_helper("chopper.proms", prom_file_names)
    func_map['audiocpu'] = helpers.name_file_helper("chopper.audiocpu", "kk3.6j")
    func_map['tx'] = helpers.name_file_helper("chopper.tx_tiles", "kk5.3a")
    bg_file_names = [
        "kk10.1a",
        "kk11.1b",
        "kk12.1d",
        "kk13.1e"
    ]
    func_map['bg'] = helpers.equal_split_helper("chopper.bg_tiles", bg_file_names)
    sp16_file_names = [
        "kk9.3g",
        "kk8.3e",
        "kk7.3d",
        "kk6.3b"
    ]
    func_map['sp16'] = helpers.equal_split_helper("chopper.sp16_tiles", sp16_file_names)
    sp32_file_names = [
        "kk18.8m",
        "kk19.8n",
        "kk20.8p",
        "kk21.8s",
        "kk14.7p",
        "kk15.7s",
        "kk16.8j",
        "kk17.8k"
    ]
    func_map['sp32'] = helpers.equal_split_helper("chopper.sp32_tiles", sp32_file_names)
    func_map['ym2'] = helpers.name_file_helper("chopper.ym2", "kk2.5b")
    func_map['plds'] = helpers.name_file_helper("chopper.plds", "p-a1.8b")
    ph_files = {
        'horizon.6h': 0x400,
        'vertical.7h': 0x400
    }
    func_map['placeholders'] = helpers.placeholder_helper(ph_files)
    out_files.append(utils.build_snk_rom("legofair.zip", bundle_contents, func_map))

    return out_files

def _handle_fantasy(bundle_contents):
    out_files = []
    # COMMON
    func_map = {}
    gfx1_file_names = [
        "fs10ic50.bin",
        "fs11ic51.bin"
    ]
    func_map['gfx1'] = helpers.equal_split_helper("fantasyu.gfx1", gfx1_file_names)
    snk6502_file_names = [
        "fs_b_51.bin",
        "fs_a_52.bin",
        "fs_c_53.bin"
    ]
    func_map['snk6502'] = helpers.equal_split_helper("fantasyu.snk6502", snk6502_file_names)
    speech_file_names = [
        "fs_d_7.bin",
        "fs_e_8.bin",
        "fs_f_11.bin"
    ]
    func_map['speech'] = helpers.equal_split_helper("fantasyu.speech", speech_file_names)
    logger.info("Processing fantasy common files...")
    common_file_map = helpers.process_rom_files(bundle_contents, func_map)

    # FANTASYJ
    func_map = {}
    prom_file_names = [
        "prom-8.bpr",
        "prom-7.bpr"
    ]
    func_map['prom'] = helpers.equal_split_helper("fantasyj.proms", prom_file_names)
    maincpu_file_names = [
        "fs5jic12.bin",
        "fs1jic7.bin",
        "fs2jic8.bin",
        "fs3jic9.bin",
        "fs4jic10.bin",
        "fs6jic14.bin",
        "fs7jic15.bin",
        "fs8jic16.bin",
        "fs9jic17.bin"
    ]
    func_map['maincpu'] = helpers.equal_split_helper("fantasyj.maincpu", maincpu_file_names)
    func_map['common'] = helpers.existing_files_helper(common_file_map)
    out_files.append(utils.build_snk_rom("fantasyj.zip", bundle_contents, func_map))

    # FANTASYU
    func_map = {}
    prom_file_names = [
        "fantasy.ic7",
        "fantasy.ic6"
    ]
    func_map['prom'] = helpers.equal_split_helper("fantasyu.proms", prom_file_names)
    maincpu_file_names = [
        "ic12.cpu",
        "ic07.cpu",
        "ic08.cpu",
        "ic09.cpu",
        "ic10.cpu",
        "ic14.cpu",
        "ic15.cpu",
        "ic16.cpu",
        "ic17.cpu"
    ]
    func_map['maincpu'] = helpers.equal_split_helper("fantasyu.maincpu", maincpu_file_names)

    func_map['common'] = helpers.existing_files_helper(common_file_map)
    out_files.append(utils.build_snk_rom("fantasyu.zip", bundle_contents, func_map))

    return out_files

def _handle_timesoldiers(bundle_contents):
    out_files = []

    # Common
    func_map = {}
    z80_file_names = [
        "bf.7",
        "bf.8",
        "bf.9"
    ]
    func_map['z80'] = helpers.equal_split_helper("TimeSoldiers.z80", z80_file_names)
    gfx2_file_names = [
        "bf.10",
        "bf.14",
        "bf.18",
        "bf.11",
        "bf.15",
        "bf.19",
        "bf.12",
        "bf.16",
        "bf.20",
        "bf.13",
        "bf.17",
        "bf.21"
    ]
    def timesold_gfx2(in_file_name, filenames):
        def process(in_files):
            contents = in_files[in_file_name]
            contents = transforms.splice_out(contents, start=15*131072, length=131072)
            contents = transforms.splice_out(contents, start=11*131072, length=131072)
            contents = transforms.splice_out(contents, start=7*131072, length=131072)
            contents = transforms.splice_out(contents, start=3*131072, length=131072)
            chunks = transforms.equal_split(contents, len(filenames))
            return dict(zip(filenames, chunks))
        return process
    func_map['gfx2'] = timesold_gfx2("TimeSoldiers.gfx2.rom", gfx2_file_names)

    logger.info("Processing timesold common files...")
    common_file_map = helpers.process_rom_files(bundle_contents, func_map)


    def timesold_maincpu_common(in_file_name, filenames):
        def maincpu(in_files):
            contents = in_files[in_file_name]
            chunks = transforms.equal_split(contents, 2)
            chunks = transforms.deinterleave_all(chunks, num_ways=2, word_size=1)
            return dict(zip(filenames, chunks))
        return maincpu

    # TIMESOLD
    func_map = {}
    maincpu_file_names = [
        "bf.3",
        "bf.4",
        "bf.1",
        "bf.2"
    ]
    func_map['maincpu'] = timesold_maincpu_common("TimeSoldiers.3.68k", maincpu_file_names)
    gfx1_file_names = [
        "bf.6",
        "bf.5"
    ]
    func_map['gfx1'] = helpers.deinterleave_helper("TimeSoldiers.u.gfx1.rom",
        gfx1_file_names, num_ways=2, word_size=1)
    func_map['common'] = helpers.existing_files_helper(common_file_map)
    out_files.append(utils.build_snk_rom("timesold.zip", bundle_contents, func_map))

    # BTLFIELD
    func_map = {}
    maincpu_file_names = [
        "bfv1_03.bin",
        "bfv1_04.bin",
        "bf.1",
        "bf.2"
    ]
    func_map['maincpu'] = timesold_maincpu_common("TimeSoldiers.j.68k", maincpu_file_names)
    gfx1_file_names = [
        "bfv1_06.bin",
        "bfv1_05.bin"
    ]
    func_map['gfx1'] = helpers.deinterleave_helper("TimeSoldiers.j.gfx1.rom",
        gfx1_file_names, num_ways=2, word_size=1)
    func_map['common'] = helpers.existing_files_helper(common_file_map)
    out_files.append(utils.build_snk_rom("btlfield.zip", bundle_contents, func_map))

    return out_files

def _handle_munchmobile(bundle_contents):
    out_files = []

    # Common
    func_map = {}
    func_map['audiocpu'] = helpers.name_file_helper("joyfulr.audiocpu", "mu.2j")
    func_map['proms'] = helpers.name_file_helper("joyfulr.proms", "a2001.clr")
    gfx1_file_names = [
        "s1.10a",
        "s2.10b"
    ]
    func_map['gfx1'] = helpers.equal_split_helper("joyfulr.gfx1", gfx1_file_names)
    gfx2_file_names = [
        "b1.2c",
        "b2.2b"
    ]
    func_map['gfx2'] = helpers.equal_split_helper("joyfulr.gfx2", gfx2_file_names)
    func_map['gfx4'] = helpers.name_file_helper("joyfulr.gfx4", "h")

    logger.info("Processing joyfulr common files...")
    common_file_map = helpers.process_rom_files(bundle_contents, func_map)

    # JOYFULR
    func_map = {}
    maincpu_file_names = [
        "m1j.10e",
        "m2j.10d"
    ]
    func_map['maincpu'] = helpers.equal_split_helper("joyfulr.maincpu", maincpu_file_names)
    gfx3_file_names = [
        "f1j.1g",
        "f2j.3g",
        "f3j.5g"
    ]
    func_map['gfx3'] = helpers.equal_split_helper("joyfulr.gfx3", gfx3_file_names)
    func_map['common'] = helpers.existing_files_helper(common_file_map)
    out_files.append(utils.build_snk_rom("joyfulr.zip", bundle_contents, func_map))

    # MNCHMOBL
    func_map = {}
    maincpu_file_names = [
        "m1.10e",
        "m2.10d"
    ]
    func_map['maincpu'] = helpers.equal_split_helper("mnchmobl.maincpu", maincpu_file_names)
    gfx3_file_names = [
        "f1.1g",
        "f2.3g",
        "f3.5g"
    ]
    func_map['gfx3'] = helpers.equal_split_helper("mnchmobl.gfx3", gfx3_file_names)
    func_map['common'] = helpers.existing_files_helper(common_file_map)
    out_files.append(utils.build_snk_rom("mnchmobl.zip", bundle_contents, func_map))

    return out_files

def _handle_sasuke(bundle_contents):
    out_files = []

    func_map = {}
    maincpu_file_names = [
        "sc1",
        "sc2",
        "sc3",
        "sc4",
        "sc5",
        "sc6",
        "sc7",
        "sc8",
        "sc9",
        "sc10"
    ]
    func_map['maincpu'] = helpers.equal_split_helper('sasuke.maincpu', maincpu_file_names)
    gfx1_file_names = [
        "mcs_c",
        "mcs_d"
    ]
    func_map['gfx1'] = helpers.equal_split_helper('sasuke.gfx1', gfx1_file_names)
    func_map['proms'] = helpers.name_file_helper("sasuke.proms", "sasuke.clr")
    func_map['snk6502'] = helpers.name_file_helper("sasuke.snk6502", "sc11")
    out_files.append(utils.build_snk_rom("sasuke.zip", bundle_contents, func_map))

    return out_files

def _handle_ozmawars(bundle_contents):
    out_files = []

    func_map = {}
    maincpu_file_names = [
        "mw01",
        "mw02",
        "mw03",
        "mw04",
        "mw05",
        "mw06"
    ]
    def ozmawars_maincpu(in_file_name, filenames):
        def maincpu(in_files):
            contents = in_files[in_file_name]
            contents = transforms.splice_out(contents, start = 8192, end = 16384)
            chunks = transforms.equal_split(contents, len(filenames))
            return dict(zip(filenames, chunks))
        return maincpu
    func_map['gfx2'] = ozmawars_maincpu("ozmawars.maincpu", maincpu_file_names)
    prom_file_names = [
        "01.1",
        "02.2"
    ]
    func_map['prom'] = helpers.equal_split_helper('moonbase.proms', prom_file_names)
    out_files.append(utils.build_snk_rom("ozmawars.zip", bundle_contents, func_map))

    return out_files

def _handle_paddlemania(bundle_contents):
    func_map = {}
    out_files = []

    prom_file_map = {
        "padlem.a": 0x100,
        "padlem.b": 0x100,
        "padlem.c": 0x100,
        "padlem.17j": 0x400,
        "padlem.16j": 0x400
    }
    func_map['prom'] = helpers.custom_split_helper('PaddleMania.proms.rom', prom_file_map)
    func_map['audiocpu'] = helpers.name_file_helper('PaddleMania.z80', 'padlem.18c')
    func_map['color_prom'] = helpers.name_file_helper('PaddleMania.clut.rom', 'padlem.18n')
    maincpu_file_names = [
        "padlem.6g",
        "padlem.3g",
        "padlem.6h",
        "padlem.3h"
    ]
    def padlem_maincpu(in_file_name, filenames):
        def maincpu(in_files):
            contents = in_files[in_file_name]
            chunks = transforms.equal_split(contents, len(filenames) // 2)
            chunks = transforms.deinterleave_all(chunks, num_ways=2, word_size=1)
            return dict(zip(filenames, chunks))
        return maincpu
    func_map['maincpu'] = padlem_maincpu('PaddleMania.1.68k', maincpu_file_names)
    gfx1_file_names = [
        "padlem.9m",
        "padlem.16m",
        "padlem.9n",
        "padlem.16n",
        "padlem.6m",
        "padlem.13m",
        "padlem.6n",
        "padlem.13n"
    ]
    func_map['gfx1'] = padlem_maincpu('PaddleMania.gfx1.rom', gfx1_file_names)
    out_files.append(utils.build_snk_rom("paddlema.zip", bundle_contents, func_map))

    return out_files


def _handle_bermuda(bundle_contents):
    '''Extract Bermuda Triangle / World Wars'''
    out_files = []

    # Bermuda Triangle Common
    func_map = {}
    sp_filenames = [
        "p9.3b",
        "p8.3d",
        "p7.3e",
        "p6.3g"
    ]
    func_map['sp'] = helpers.equal_split_helper('BermudaTriangle.sp', sp_filenames)
    sp32_filenames = [
        "p11.7p",
        "p12.7s",
        "p13.8h",
        "p14.8k",
        "p15.8m",
        "p16.8n",
        "p17.8p",
        "p18.8s"
    ]
    func_map['sp32'] = helpers.equal_split_helper('BermudaTriangle.sp32', sp32_filenames)
    bg_filenames = [
        "p19.1", 
        "p20.1b",
        "p21.1d",
        "p22.1e"
    ]
    func_map['bg'] = helpers.equal_split_helper('BermudaTriangle.bg', bg_filenames)
    pal_filenames = [
        "2.1l",
        "1.1k",
        "3.2l"
    ]
    func_map['pal'] = utils.simple_palette_helper('BermudaTriangle.pal', pal_filenames)
    ph_files = {
        'horizon.6h': 0x400,
        'vertical.7h': 0x400
    }
    func_map['tx'] = helpers.name_file_helper("BermudaTriangle.tx", "p10.3a")
    func_map['sub'] = helpers.name_file_helper("BermudaTriangle.1.z80", "p2.8p")
    func_map['ph'] = helpers.placeholder_helper(ph_files)
    logger.info("Processing Bermudat Triangle common files...")
    common_file_map = helpers.process_rom_files(bundle_contents, func_map)

    # BERMUDAT
    func_map = {}
    func_map['maincpu'] = helpers.name_file_helper("BermudaTriangle.0.z80", "p1.4p")
    func_map['audiocpu'] = helpers.name_file_helper("BermudaTriangle.2.z80", "p3.7k")
    adpcm_filenames = [
        "p4.5e",
        "p5.5g"
    ]
    func_map['adpcm'] = helpers.equal_split_helper('BermudaTriangle.adpcm', adpcm_filenames)
    func_map['common'] = helpers.existing_files_helper(common_file_map)
    mame_name = "bermudat.zip"
    logger.info(f"Building {mame_name}...")
    out_files.append(
        {'filename': mame_name, 'contents': helpers.build_rom(bundle_contents, func_map)}
    )
    logger.info(f"Extracted {mame_name}.")


    # BERMUDATJ
    func_map = {}
    func_map['maincpu'] = helpers.name_file_helper("BermudaTriangle.j.0.z80", "p1.4p")
    func_map['audiocpu'] = helpers.name_file_helper("BermudaTriangle.j.2.z80", "p3.7k")
    adpcm_filenames = [
        "p4.5e",
        "p5.5g"
    ]
    func_map['adpcm'] = helpers.equal_split_helper('BermudaTriangle.j.adpcm', adpcm_filenames)
    func_map['common'] = helpers.existing_files_helper(common_file_map)
    mame_name = "bermudatj.zip"
    logger.info(f"Building {mame_name}...")
    out_files.append(
        {'filename': mame_name, 'contents': helpers.build_rom(bundle_contents, func_map)}
    )
    logger.info(f"Extracted {mame_name}.")

    return out_files


def _handle_worldwars(bundle_contents):
    '''Extract World Wars'''
    out_files = []
    # World Wars Common
    func_map = {}
    bg_filenames = [
        "ww11.1e",
        "ww12.1d",
        "ww13.1b",
        "ww14.1a"
    ]
    func_map['bg'] = helpers.equal_split_helper('WorldWars.bg', bg_filenames)
    sp_filenames = [
        "ww10.3g",
        "ww9.3e",
        "ww8.3d",
        "ww7.3b"
    ]
    func_map['sp'] = helpers.equal_split_helper('WorldWars.sp', sp_filenames)
    sp32_filenames = [
        "ww21.7p",
        "ww22.7s",
        "ww19.8h",
        "ww20.8k",
        "ww15.8m",
        "ww16.8n",
        "ww17.8p",
        "ww18.8s"
    ]
    func_map['sp32'] = helpers.equal_split_helper('WorldWars.sp32', sp32_filenames)
    ph_files = {
        'horizon.5h': 0x400,
        'vertical.7h': 0x400
    }
    func_map['ph'] = helpers.placeholder_helper(ph_files)
    logger.info("Processing World Wars common files...")
    common_file_map = helpers.process_rom_files(bundle_contents, func_map)

    # WORLDWAR
    func_map = {}
    func_map['maincpu'] = helpers.name_file_helper("WorldWars.0.z80", "ww4.4p")
    func_map['sub'] = helpers.name_file_helper("WorldWars.1.z80", "ww5.8p")
    func_map['audiocpu'] = helpers.name_file_helper("WorldWars.2.z80", "ww3.7k")
    func_map['tx'] = helpers.name_file_helper("WorldWars.tx", "ww6.3a")
    func_map['common'] = helpers.existing_files_helper(common_file_map)
    pal_filenames = [
        "2.1l",
        "1.1k",
        "3.2l"
    ]
    func_map['pal'] = utils.simple_palette_helper('WorldWars.pal', pal_filenames)
    adpcm_filenames = [
        "p4.5e",
        "p5.5g"
    ]
    func_map['adpcm'] = helpers.equal_split_helper('WorldWars.adpcm', adpcm_filenames)
    mame_name = "worldwar.zip"
    logger.info(f"Building {mame_name}...")
    out_files.append(
        {'filename': mame_name, 'contents': helpers.build_rom(bundle_contents, func_map)}
    )
    logger.info(f"Extracted {mame_name}.")

    # BERMUDATA - This is actually World Wars (J)!
    func_map = {}
    func_map['maincpu'] = helpers.name_file_helper("WorldWars.j.0.z80", "wwu4.4p")
    func_map['sub'] = helpers.name_file_helper("WorldWars.j.1.z80", "wwu5.8p")
    func_map['audiocpu'] = helpers.name_file_helper("WorldWars.j.2.z80", "wwu3.7k")
    func_map['tx'] = helpers.name_file_helper("WorldWars.j.tx", "wwu6.3a")
    func_map['common'] = helpers.existing_files_helper(common_file_map)
    pal_filenames = [
        "u2bt.2l",
        "u1bt.1k",
        "u3bt.1l"
    ]
    func_map['pal'] = utils.simple_palette_helper('WorldWars.j.pal', pal_filenames)
    out_files.append(utils.build_snk_rom("bermudata.zip", bundle_contents, func_map))

    return out_files

def _handle_marvin(bundle_contents):
    '''Extract Marvin's Maze'''
    out_files = []
    func_map = {}
    maincpu_filenames = [
        "pa1",
        "pa2",
        "pa3"
    ]
    func_map['maincpu'] = helpers.equal_split_helper('MarvinsMaze.0.z80', maincpu_filenames)
    func_map['sub'] = helpers.name_file_helper("MarvinsMaze.1.z80", "pb1")
    audiocpu_filenames = [
        "m1",
        "m2"
    ]
    func_map['audiocpu'] = helpers.equal_split_helper('MarvinsMaze.2.z80', audiocpu_filenames)
    func_map['bg'] = helpers.name_file_helper("MarvinsMaze.bg", "b2")
    func_map['fg'] = helpers.name_file_helper("MarvinsMaze.fg", "b1")
    sp_filenames = [
        "f1",
        "f2",
        "f3"
    ]
    func_map['sp'] = helpers.equal_split_helper('MarvinsMaze.sp', sp_filenames)
    func_map['tx'] = helpers.name_file_helper("MarvinsMaze.tx", "s1")
    pals_filenames = [
        "marvmaze.j2",
        "marvmaze.j1",
        "marvmaze.j3"
    ]
    func_map['pal'] = utils.palette_rebuild_helper(pals_filenames, 'MarvinsMaze.pal')
    out_files.append(utils.build_snk_rom("marvins.zip", bundle_contents, func_map))

    return out_files



'''Utility Functions specifically for SNK40'''
import logging
from gex.lib.tasks import helpers
from gex.lib.utils.blob import transforms

logger = logging.getLogger('gextoolbox')

def build_snk_rom(mame_name, bundle_contents, func_map):
    '''Helper to compact the logging output and build call'''
    logger.info(f"Building {mame_name}...")
    out_file = {'filename': mame_name, 'contents': helpers.build_rom(bundle_contents, func_map)}
    logger.info(f"Extracted {mame_name}.")
    return out_file

def simple_palette_helper(in_file_ref, pal_filenames):
    '''Rebuild RGB Palette ROMs'''
    def palette(in_files):
        in_data = in_files[in_file_ref]
        pal_contents = transforms.deinterleave_nibble(in_data, 4)
        del pal_contents[2] # Remove the spacing entry
        return dict(zip(pal_filenames, pal_contents))
    return palette

def palette_rebuild_helper(file_metas, in_file_name):
    '''Helper to use generate_snk_palette_file to bit reorder a shuffled palette'''
    def palette(in_files):
        contents = in_files[in_file_name]
        return generate_snk_palette_file(file_metas, contents)
    return palette

def generate_snk_palette_file(filenames, contents):
    '''Given a set of CRC/SHA/size, generate a set nibble-based palette file via reordering'''

    # Apply the shuffle
    bit_shuffle_order = [1, 2, 12, 13, 4, 5, 6, 0, 8, 9, 10, 11, 7, 3, 14, 15]
    shuffled_contents = transforms.bit_shuffle(contents, 2, bit_shuffle_order)

    # Deinterleave the result
    pal_contents = transforms.deinterleave_nibble(shuffled_contents, 4)
    del pal_contents[2] # Remove the spacing entry

    return dict(zip(filenames, pal_contents))


# This is the original function used to search for the Palette reorder pattern
# As it was discovered that the shuffle order is always the same, we can do this MUCH faster.
# Keeping this function for posterity/future use
# def search_snk_palette_file(file_metas, in_data):
#     '''Given a set of CRC/SHA/size, generate a set nibble-based palette file via reordering'''

#     # move to a bit array
#     in_data_bits = bitarray()
#     in_data_bits.frombytes(in_data)

#     # get rid of 3rd nibbles
#     for i in range(len(in_data_bits), 0, -16):
#         del in_data_bits[i-8:i-4]

#     bit_order_pool = list(range(0,12))

#     out_files = {}

#     # Try each permutation (11800 permutations!)
#     for count, curr_permutation in enumerate(itertools.permutations(bit_order_pool, 4)):

#         # Build the output file
#         out_bitarray = bitarray()
#         for set_offset in range(0, len(in_data_bits), 12):
#             out_bitarray.extend(bitarray('0000'))
#             for src_bit in curr_permutation:
#                 out_bitarray.append(in_data_bits[set_offset + src_bit])

#         # Get the CRC
#         out_bytes = out_bitarray.tobytes()
#         crc = hash_helper.get_crc(out_bytes)

#         for file_meta in file_metas:
#             if (crc == file_meta['crc']):
#                 print(f"CRC HIT at {count} with {str(list(curr_permutation))}!")
#                 sha = hash_helper.get_sha1(out_bytes)
#                 if (sha == file_meta['sha']):
#                     print(f"SHA HIT at {count} with {str(list(curr_permutation))}!")
#                     out_files[file_meta['filename']] = out_bytes

#         if len(file_metas) == len(out_files):
#             break

#     return out_files



'''Implementation of mmxlc1: Mega Man X Legacy Collection 1'''
from gex.lib.tasks.splicetask import SpliceTask
from gex.lib.tasks import helpers

class MegaManXLegacyCollection1Task(SpliceTask):
    '''Implements mmxlc1: Mega Man X Legacy Collection 1'''
    _task_name = "mmxlc1"
    _title = "Mega Man X Legacy Collection 1"
    _details_markdown = '''
Based on: https://github.com/s3phir0th115/MMXLC1-Rom-Extractor/blob/master/mmxlc_rom_extract.py

Mega Man X4 does not appear to be ROM based, but investigation is ongoing.
'''
    _default_input_folder = helpers.gen_steam_app_default_folder("Mega Man X Legacy Collection")
    _input_folder_desc = "Steam MMxLC install folder"



'''Implementation of genesis: Sega Genesis and Mega Drive Collection'''
import logging
import os
from gex.lib.tasks.basetask import BaseTask
from gex.lib.tasks import helpers

logger = logging.getLogger('gextoolbox')

class GenesisTask(BaseTask):
    '''Implments genesis: Sega Genesis and Mega Drive Collection'''
    _task_name = "genesis"
    _title = "Sega Genesis and Mega Drive Collection"
    _details_markdown = '''
These are the ROMs just sitting in the uncompressed ROMs folder.
There are a few more regional variants that are only available in PAK files, but this tool doesn't have a Python SGMDC PAK extractor yet 
    '''
    _default_input_folder = helpers.gen_steam_app_default_folder("Sega Classics")
    _input_folder_desc = "Sega Classics Steam Folder"

    def get_out_file_info(self):
        '''Return a list of output files'''
        return {
            "files": self._metadata['out']['files'],
            "notes": self._metadata['out']['notes']
        }

    def execute(self, in_dir, out_dir):
        '''Copy/rename the ROM files'''
        for file_metadata in self._metadata['in']['files'].values():
            resolved_file = self.read_datafile(in_dir, file_metadata)
            if 'copy_to' in file_metadata:
                out_file_entry = [x for x in self._metadata['out']['files'] if x['game'] == file_metadata['copy_to']][0]

                filename = out_file_entry['filename']
                _ = self.verify_out_file(filename, resolved_file['contents'])
                out_path = os.path.join(out_dir, filename)
                with open(out_path, "wb") as out_file:
                    out_file.write(resolved_file['contents'])
            else:
                logger.warning("Compressed ROMs not yet supported!")
        logger.info("Processing complete.")



'''Implementation of mmbnlc2: Mega Man Battle Network Legacy Collection 2'''
import logging
import os
from gex.lib.tasks.basetask import BaseTask
from gex.lib.tasks import helpers

logger = logging.getLogger('gextoolbox')

class MegaManLegacyCollection1Task(BaseTask):
    '''Implements mmbnlc2: Mega Man Battle Network Legacy Collection 2'''
    _task_name = "mmbnlc2"
    _title = "Mega Man Battle Network Legacy Collection 2"
    _details_markdown = '''
Based on zip file discoveries from https://github.com/farmerbb/RED-Project/issues/101
'''
    _default_input_folder = helpers.gen_steam_app_default_folder("MegaMan_BattleNetwork_LegacyCollection_Vol2")
    _input_folder_desc = "'MegaMan_BattleNetwork_LegacyCollection_Vol2' Folder (Steam MMBNLC1 install folder)"

    def execute(self, in_dir, out_dir):
        bundle_contents = {}
        for file_ref, file_metadata in self._metadata['in']['files'].items():
            resolved_file = self.read_datafile(in_dir, file_metadata)
            bundle_contents[file_ref] = resolved_file['contents']

        out_files = []
        out_files.extend(
            helpers.pull_files_from_archive(bundle_contents['exe4.dat'], "zip", {
                "MegaMan Battle Network 4 Red Sun.gba": "exe4/rom_e.srl",
                "RockMan.EXE 4 Tournament Red Sun.gba": "exe4/rom.srl"
            })
        )
        out_files.extend(
            helpers.pull_files_from_archive(bundle_contents['exe4b.dat'], "zip", {
                "MegaMan Battle Network 4 Blue Moon.gba": "exe4b/rom_b_e.srl",
                "RockMan.EXE 4 Tournament Blue Moon.gba": "exe4b/rom_b.srl"
            })
        )
        out_files.extend(
            helpers.pull_files_from_archive(bundle_contents['exe5.dat'], "zip", {
                "MegaMan Battle Network 5 Team Blue.gba": "exe5/rom_e.srl",
                "RockMan.EXE 5 Tournament Team of Blues.gba": "exe5/rom.srl"
            })
        )
        out_files.extend(
            helpers.pull_files_from_archive(bundle_contents['exe5k.dat'], "zip", {
                "MegaMan Battle Network 5 Team Colonel.gba": "exe5k/rom_k_e.srl",
                "RockMan.EXE 5 Tournament Team of Colonel.gba": "exe5k/rom_k.srl"
            })
        )
        out_files.extend(
            helpers.pull_files_from_archive(bundle_contents['exe6.dat'], "zip", {
                "MegaMan Battle Network 6 Cybeast Gregar.gba": "exe6/rom_e.srl",
                "RockMan.EXE 6 Tournament Cyber Beast Glaga.gba": "exe6/rom.srl"
            })
        )
        out_files.extend(
            helpers.pull_files_from_archive(bundle_contents['exe6f.dat'], "zip", {
                "MegaMan Battle Network 6 Cybeast Falzar.gba": "exe6f/rom_f_e.srl",
                "RockMan.EXE 6 Tournament Cyber Beast Falzer.gba": "exe6f/rom_f.srl"
            })
        )

        for out_file_entry in out_files:
            filename = out_file_entry['filename']
            _ = self.verify_out_file(filename, out_file_entry ['contents'])
            out_path = os.path.join(out_dir, filename)
            with open(out_path, "wb") as out_file:
                logger.info(f"Writing {filename}...")
                out_file.write(out_file_entry['contents'])

        logger.info("Processing complete.")


'''Convenience functions for hashing a binary blob'''
import hashlib
import zlib

def get_crc(content):
    '''Get the common CRC32 hash for a blob'''
    return hex(zlib.crc32(content) & 0xffffffff)

def get_md5(content):
    '''Get the MD5 sum for a blob'''
    return hashlib.md5(content).hexdigest()

def get_sha1(content):
    '''Get the SHA-1 hash for a blob'''
    return hashlib.sha1(content).hexdigest()



import os
import logging
import click
import click_log
from gex.lib.archive import arc
from gex.lib.file import identify
from gex.lib.utils import helper

logger = logging.getLogger('gextoolbox')

@click.command()
@click.option('--in', 'in_file', help = 'path to input ARC archive', required=True)
@click.option('--out', 'out_dir', help = 'path to output folder (must exist)', required=True)
@click_log.simple_verbosity_option(logger)
def extract(in_file, out_dir):
    """Extract the contents of a ARC archive"""
    try:
        with open(in_file, "rb") as curr_file:
            file_content = bytearray(curr_file.read())
            arc_contents = arc.extract(file_content)

            # Extract files
            for _, file_entry in arc_contents.items():
                contents = file_entry['contents']

                try:
                    type_id = identify.enhanced_magic_from_buffer(contents)
                    if not type_id == identify.ARC:
                        logger.warning(f'Found {type_id} file, will try to extract anyway...')
                except Exception as _:
                    logger.warning('Cannot typecheck!')

                out_file_path = os.path.join(
                    helper.cleanpath(out_dir),
                    file_entry['path'].replace("\\", os.sep)
                )
                os.makedirs(os.path.dirname(out_file_path), exist_ok = True)
                with open(out_file_path, "wb") as out_file:
                    out_file.write(contents)

        logger.info('Extraction complete.')
    except Exception as error:
        logger.error(error)
        logger.error('Error While Opening the file!')



'''
Module to decompress SEGA .prs files

SEGA PRS Decompression (LZS variant)
From: https://forums.qhimm.com/index.php?topic=11225.0 (Micky)
Credits: based on information/comparing output with
  Nemesis/http://www.romhacking.net/utils/671/
  puyotools/http://code.google.com/p/puyotools/
  fuzziqer software prs/http://www.fuzziqersoftware.com/projects.php

# TODO: This was quickly converted from python2, and can likely be optimized/made more pythonic
'''
import array

class DecompressPrs:
    '''Class to decompress a SEGA .prs file (LZ77)'''
    def __init__(self, data):
        self.ibuf = array.array("B", data)
        self.obuf = array.array("B")

        self.iofs = 0
        self.bit = 0
        self.cmd = 0

    def _get_byte(self):
        val = self.ibuf[self.iofs]
        self.iofs += 1
        return val

    def _get_bit(self):
        if self.bit == 0:
            self.cmd = self._get_byte()
            self.bit = 8
        bit = self.cmd & 1
        self.cmd >>= 1
        self.bit -= 1
        return bit

    def decompress(self):
        '''Decompress the PRS file this class was created around'''
        while self.iofs < len(self.ibuf):
            cmd = self._get_bit()
            if cmd:
                self.obuf.append(self.ibuf[self.iofs])
                self.iofs += 1
            else:
                t = self._get_bit()
                if t:
                    byte_a = self._get_byte()
                    byte_b = self._get_byte()

                    offset = ((byte_b << 8) | byte_a) >> 3
                    amount = byte_a & 7
                    if self.iofs < len(self.ibuf):
                        if amount == 0:
                            amount = self._get_byte() + 1
                        else:
                            amount += 2

                    start = len(self.obuf) - 0x2000 + offset
                else:
                    amount = 0
                    for _ in range(2):
                        amount <<= 1
                        amount |= self._get_bit()
                    offset = self._get_byte()
                    amount += 2

                    start = len(self.obuf) - 0x100 + offset
                for _ in range(amount):
                    if start < 0:
                        self.obuf.append(0)
                    elif start < len(self.obuf):
                        self.obuf.append(self.obuf[start])
                    else:
                        self.obuf.append(0)
                    start += 1

        return self.obuf.tobytes()



'''Implementation of bubsy: Bubsy Two-Fur'''
from gex.lib.tasks.copytask import CopyTask
from gex.lib.tasks import helpers

class BubsyTask(CopyTask):
    '''Implements bubsy: Bubsy Two-Fur'''
    _task_name = "bubsy"
    _title = "Bubsy Two-Fur"
    _details_markdown = '''
These are the ROMs just sitting in the install folder. 
    '''
    _default_input_folder = helpers.gen_steam_app_default_folder("Bubsy Two-Fur")
    _input_folder_desc = "Bubsy Two-Fur Steam Folder"



'''Implemntation of disneyac: Disney Afternoon Collection'''
from gex.lib.tasks.splicetask import SpliceTask
from gex.lib.tasks import helpers

class DisneyAfternoonCollectionTask(SpliceTask):
    '''Implemnts disneyac: Disney Afternoon Collection'''
    _task_name = "disneyac"
    _title = "Disney Afternoon Collection"
    _details_markdown = '''
Based on MMLC & DAC Extractor - https://github.com/HTV04/mmlc-dac-extractor
'''
    _out_file_notes = {}
    _default_input_folder = helpers.gen_steam_app_default_folder("DisneyAfternoon")
    _input_folder_desc = "DisneyAfternoon Folder (Steam install folder)"



'''Module to handle ZIP archive (additional utility)'''
import io
import zipfile
import logging

logger = logging.getLogger('gextoolbox')

def extract(bytes_obj):
    '''Extract a ZIP archive into a list'''
    files = {}
    with zipfile.ZipFile(io.BytesIO(bytes_obj), "r") as archive:
        zip_entries = list(archive.infolist())
        for file_entry in zip_entries:
            with archive.open(file_entry) as file_read_obj:
                contents = file_read_obj.read()
                files[file_entry.filename] = {
                    "contents": contents,
                    "entry": file_entry.filename,
                    "size": file_entry.file_size
                }
    return files

def get_metadata(bytes_obj):
    '''Get a list of file entries usable for verifying the contents'''
    files = {}
    with zipfile.ZipFile(io.BytesIO(bytes_obj), "r") as archive:
        zip_entries = list(archive.infolist())
        for file_entry in zip_entries:
            files[file_entry.filename] = {
                "filename": file_entry.filename,
                "size": file_entry.file_size,
                "crc": hex(file_entry.CRC)[2:].upper().rjust(8, "0")
            }
    return files



'''INCOMPLETE Implementation of cas1: Capcom Arcade Stadium 1'''
import glob
import logging
import os

from gex.lib.archive import kpka
from gex.lib.tasks import helpers
from gex.lib.tasks.basetask import BaseTask

logger = logging.getLogger('gextoolbox')

# TODO: Further research required

class CAS1Task(BaseTask):
    '''INCOMPLETE Implements cas1: Capcom Arcade Stadium 1'''
    _task_name = 'cas1'
    _title = "Capcom Arcade Stadium 1"
    _details_markdown = '''
## INCOMPLETE!

This script is a placeholder for full CAS1 extraction. See https://github.com/shawngmc/game-extraction-toolbox/issues/18 for latest details!

Currently, it extracts nothing; however, with a "--prop include-partials=True", it will extract all KPKA paks with the following naming convention:
[ARCHIVEID]_[SEQUENTIALNUM]_[SIZE].dat
These are not playable ROMs!

## Analysis
### KPKA
These appear to be KPKA archives, and it is reasonably certain that the extraction is correct, as most files have a 3-4 character allcaps header
indicating file type. Note that the file order can differ between KPKA files; the size and header bytes are best used for identification.

### File Types
It's more effective to actually categorize the files.

#### ROM Data (New Only)
 - Has no header
 - Large (very similar to the extracted ZIP file sizes, but slightly smaller - perhaps diff compression)
 - 1-2 per PAK (typically 2, but only 1 if the game didn't have a US release)
 - Is this an encrypted version of the PKZIP file?

#### Very Small MMAC
 - MMAC Header
 - Always 26 bytes
 - 1-2 per PAK (does not necessarily match the number of ROM Data packages)

#### Moderate MMAC
 - MMAC Header
 - Typically a couple K in size

#### 'MAME' file
 - MAME header
 - Typically less than 1K
 - Some internal paths, some also mention DSW1/DSW2 - could these be DIP Switches?
 - Like a MAME engine config file for the ROM, to tell it MAME driver name, etc.?

#### SCN
 - SCN Header
 - Typically less than 1K
 - RE Engine Scene Description?
 - Has paths for other related files

#### PSB
 - '.psb' header
 - Always 331 bytes
 - CAS1 - all PSBs are the same
 - CAS2 - no per-pak PSB, 1 in the main archive
 - CAS1 and CAS2 a little different
   - 0x31-0x35 - 5 bytes
   - 0x3A-0x46 - 13 bytes
   - 0x6F-0x82 - 20 bytes
   - 38/331 bytes - about 11%

#### PKZIP (Old Depot Only)
 - 'PK' Header
 - Appears to be a standard ZIP file with the ROM

##### Diff vs. M2 PSB
In the example above, 1556708_1_331.dat appears to have ".psb" at the beginning; however, this doesn't match what we know about the PSB filetype.
 - If this was an uncompressed PSB file, we would expect the first three bytes to be "PSB"
 - If this were a compressed PSB file, we would expect the first three bytes to be a compression type indicator, like "mdf"
 - We have no current indication M2 (which makes the m2engage emulator and MArchive file format) worked on CAS
 - MArchiveBatchTool can't seem to do anything with the file.
 - The file is far shorter than most PSB files.
As such, 'PSB' is likely a coincidence or other dead end.
For CAS1, every '.psb' file is identical. CAS2 does not have these '.psb' files (at least in the per-DLC packages). 
However, there is a '.psb' file in the main package (look for the 331 byte file size/file 1888).
It is definitely different than the CAS1 PSB, but only about 30% of the file is changed. (0x30-0x47, 0x6F-0x81)

### Examples
#### Example 1 -  (1556708)
 - 1556708_0_1957557.dat - One of the ROM packages
 - 1556708_1_331.dat - has '.psb' at the beginning, but does not appear to be a valid M2 MArchive/PSB Archive
 - 1556708_2_1976307.dat - One of the ROM packages
 - 1556708_3_150.dat - Small MAME file describing ROM name?
 - 1556708_4_138.dat - Small file of null bytes
 - 1556708_5_667.dat - SCN file - Scene description?
 - 1556708_6_3828.dat - Moderate MMAC file listing ROM file names and possible keys?
 - 1556708_7_26.dat - Very small MMAC file

#### Example 2 - 1556700 (Vulgus)
| Old                | New                 | Info                                                            | 
|--------------------|---------------------|-----------------------------------------------------------------| 
| 1_346.dat          | 1_346.dat           | MAME file; both 346 bytes, identical, for 'vulgus', has DSWs    | 
| 6_134.dat          | 7_134.dat           | MAME file; both 134 bytes, identical, for 'vulgusj'             | 
| 4_26.dat           | 5_26.dat            | Very Small MMAC; Both 26 bytes, 1 byte diff (0x08)              | 
| 0_14.dat           | 0_26.dat            | Very Small MMAC; 14 vs 26 bytes                                 | 
| 5_655.dat          | 6_655.dat           | SCN file; Both 655 bytes, identical                             | 
| N/A                | 3_331.dat           | '.psb' file; only in new                                        | 
| 2_71436.dat        | 2_70421.dat         | PK vs ??? file, guessing via size and orig this is vulgusj      | 
| 3_71071.dat        | 4_70123.dat         | PK vs ??? file, guessing via size and orig this is vulgus       | 
    '''
    _out_file_list = [
    ]

    _out_file_notes = {
    }
    _default_input_folder = helpers.gen_steam_app_default_folder("Capcom Arcade Stadium")
    _input_folder_desc = "CAS 1 Folder"
    _prop_info = {
        "include-partials": {
            "description": "Include the partial ROMs that are missing data; useful for mixing with other sources or investigation",
            "default": False,
            "type": "Boolean"
        }
    }

    def _find_files(self, in_path):
        return glob.glob(os.path.join(in_path, "**", "*.pak"), recursive=True)

    def execute(self, in_dir, out_dir):
        
        if self._props.get('include-partials'):
            pak_files = self._find_files(in_dir)
            for file in pak_files:
                file_id = None
                filename = os.path.basename(file)
                if "patch" in filename:
                    # logger.info(f"Skipping patch file {filename}")
                    file_id = "patch"
                elif filename == "re_chunk_000.pak":
                    file_id = "chunk"
                else:
                    file_id = file.split(".")[0].split("_")[-1]

                logger.info(f"Extracting {file}: {file_id}")
                try:
                    with open(file, "rb") as curr_file:
                        file_content = bytearray(curr_file.read())
                        kpka_contents = kpka.extract(file_content)
                        output_files = []

                        for kpka_contents in kpka_contents.values():
                            print(f"{kpka_contents['entry']}: {kpka_contents['size']}")
                            output_files.append({'filename': f"{file_id}_{kpka_contents['entry']}_{kpka_contents['size']}.dat", "contents": kpka_contents['contents']})

                        for output_file in output_files:
                            out_path = os.path.join(out_dir, output_file['filename'])
                            with open(out_path, "wb") as out_file:
                                out_file.write(output_file['contents'])
                except OSError as error:
                    logger.warning(f'Error while processing {file}!')
                    logger.warning(error)
        else:
            logger.error("CAS1 is not yet implemented; to help investigate, add --prop 'include-partials=True' to extract what we have so far!")

        logger.info("Processing complete.")



'''Module to handle ARC archives (Capcom MT Engine)'''
import zlib
import logging

def extract(bytes_obj):
    '''Extract an ARC archive (Capcom MT Engine)'''
    magic_string = bytes_obj[0:3].decode("utf-8")
    if magic_string != "ARC":
        raise Exception("Not a valid ARC archive!")
    else:
        files = {}
        logging.debug("ARC detected!")
        logging.debug(f'Archive Size: {len(bytes_obj)}')
        version = int.from_bytes(bytes_obj[4:6], byteorder='little')
        logging.debug(f'Version: {version}')
        total_files = int.from_bytes(bytes_obj[6:8], byteorder='little')
        logging.debug(f'Number of files: {total_files}')

        curr_pos = 0x8
        for curr_file in range(0, total_files):
            logging.debug(f'  file: {curr_file}')
            path = str(bytes_obj[curr_pos:curr_pos+0x40], 'utf-8').rstrip('\x00')
            logging.debug(f'    path: {path}')
            comp_type = int.from_bytes(bytes_obj[curr_pos+0x40:curr_pos+0x44], byteorder='little')
            logging.debug(f'    type: {comp_type}')
            zsize = int.from_bytes(bytes_obj[curr_pos+0x44:curr_pos+0x48], byteorder='little')
            logging.debug(f'    zsize: {zsize}')
            size = int.from_bytes(bytes_obj[curr_pos+0x48:curr_pos+0x4b], byteorder='little')
            logging.debug(f'    size: {size}')
            offset = int.from_bytes(bytes_obj[curr_pos+0x4c:curr_pos+0x50], byteorder='little')
            logging.debug(f'    offset: {offset}')

            contents = bytes_obj[offset:offset+zsize]
            contents = zlib.decompress(contents, wbits = 0)
            logging.debug(f'    decompressed size: {len(contents)} bytes')

            files[curr_file] = {
                "contents": contents,
                "offset": offset,
                "entry": curr_file,
                "size": size,
                "path": path
            }
            curr_pos = curr_pos + 0x50

        return files



'''Implementation of psikyo: Psikyo Shooter Collector's Bundle'''
import glob
import logging
import os
from gex.lib.utils.blob import transforms
from gex.lib.tasks import helpers
from gex.lib.tasks.basetask import BaseTask

logger = logging.getLogger('gextoolbox')

class PsikyoTask(BaseTask):
    '''Implements psikyo: Psikyo Shooter Collector's Bundle'''
    _task_name = "psikyo"
    _title = "Psikyo Shooter Collector's Bundle"
    _details_markdown = '''
'''
    _game_info_list = [
        {
            "game": "Dragon Blaze",
            "in": {
                "base_dir": "Dragon Blaze\\Data"
            },
            "filename": "dragnblz.zip",
            "status": "partial",
            "handler": "_handle_dragnblz",
            "notes": []
        },
        {
            "game": "Strikers 1945",
            "in": {
                "base_dir": "STRIKERS1945\\Data"
            },
            "filename": "s1945.zip",
            "status": "partial",
            "handler": "_handle_copyorig",
            "notes": []
        },
        {
            "game": "Strikers 1945 II",
            "in": {
                "base_dir": "STRIKERS1945 Ⅱ\\Data"
            },
            "filename": "s1945ii.zip",
            "status": "partial",
            "handler": "_handle_copyorig",
            "notes": []
        },
        {
            "game": "Strikers 1945 III",
            "in": {
                "base_dir": "STRIKERS1945 III\\Data"
            },
            "filename": "s1945iii.zip",
            "status": "partial",
            "handler": "_handle_copyorig",
            "notes": []
        },
        {
            "game": "GUNBARICH",
            "in": {
                "base_dir": "GUNBARICH\\Data"
            },
            "filename": "gnbarich.zip",
            "status": "partial",
            "handler": "_handle_copyorig",
            "notes": []
        },
        {
            "game": "GUNBIRD",
            "in": {
                "base_dir": "GUNBIRD\\Data"
            },
            "filename": "gunbird.zip",
            "status": "partial",
            "handler": "_handle_copyorig",
            "notes": []
        },
        {
            "game": "GUNBIRD 2",
            "in": {
                "base_dir": "GUNBIRD2\\Data"
            },
            "filename": "gunbird2.zip",
            "status": "partial",
            "handler": "_handle_copyorig",
            "notes": []
        },
        {
            "game": "SOL DIVIDE -SWORD OF DARKNESS-",
            "in": {
                "base_dir": "SOL DIVIDE -SWORD OF DARKNESS-\\Data"
            },
            "filename": "soldivid.zip",
            "status": "partial",
            "handler": "_handle_copyorig",
            "notes": []
        },
        {
            "game": "ZERO GUNNER 2",
            "in": {
                "base_dir": "ZERO GUNNER 2-\\Data"
            },
            "filename": "zerogu2.zip",
            "status": "partial",
            "handler": "_handle_copyorig",
            "notes": []
        },
        {
            "game": "TENGAI",
            "in": {
                "base_dir": "TENGAI\\Data"
            },
            "filename": "tengai.zip",
            "status": "partial",
            "handler": "_handle_copyorig",
            "notes": []
        },
        {
            "game": "Samurai Aces",
            "in": {
                "base_dir": "Samurai Aces\\Data"
            },
            "filename": "samuraia.zip",
            "status": "partial",
            "handler": "_handle_copyorig",
            "notes": []
        },
        {
            "game": "Samurai Aces III: Sengoku Cannon",
            "in": {
                "base_dir": "SENGOKU CANNON\\Data"
            },
            "filename": "sngkace.zip",
            "status": "partial",
            "handler": "_handle_copyorig",
            "notes": []
        }
    ]
    _out_file_notes = {}
    _default_input_folder = helpers.STEAM_APP_ROOT
    _input_folder_desc = "Steam library root"

    def __init__(self):
        super().__init__()
        self._out_file_list = map(lambda x: {
            'filename': x['filename'],
            'game': x['game'],
            'status': x['status'],
            'system': "Arcade",
            "notes": x['notes']},
            self._game_info_list)

    def _read_all_files(self, base_path):
        file_paths = glob.glob(base_path + '\\**\\*.*', recursive=True)
        files = {}
        for file_path in file_paths:
            with open(file_path, 'rb') as file_obj:
                file_data = file_obj.read()
                files[os.path.basename(file_path)] = file_data
        return files

    def execute(self, in_dir, out_dir):
        output_files = []
        for game in self._game_info_list:
            if game.get('status') == 'partial' and not self._props.get('include-partials'):
                logger.info(f"Skipping {game['game']} as complete extraction isn't possible.")
                continue

            game_dir = os.path.join(in_dir, game['in']['base_dir'])
            if os.path.exists(game_dir):
                logger.info(f"Extracting {game['game']}...")

                # Load the files
                in_files = self._read_all_files(game_dir)
                # in_files = {}
                # for in_path_fragment in game['in']['files']:
                #     in_path = os.path.join(game_dir, in_path_fragment)
                #     with open(in_path, "rb") as curr_file:
                #         in_files[in_path_fragment] = curr_file.read()

                if 'handler' in game:
                    handler_func = getattr(self, game['handler'])
                    output_files.append(handler_func(in_files, game))
                else:
                    logger.warning(f"No handler defined for {game['game']}")

        for output_file in output_files:
            with open(os.path.join(out_dir, output_file['filename']), "wb") as out_file:
                out_file.write(output_file['contents'])

        logger.info("Processing complete.")


    def _handle_dragnblz(self, in_files, game_info):
        func_map = {}

        # Temporary working
        def raw_files(in_files): 
            return {f'orig/{k}':v for (k,v) in in_files.items()}
        func_map["raw"] = raw_files

        # maincpu - identical, but 426 bytes missing at the end of each file
        def maincpu(in_files): 
            contents = in_files['SH922.BIN']
            contents = transforms.swap_endian(contents)
            chunks = transforms.deinterleave(contents, num_ways=2, word_size=2)
            filenames = [
                "2prog_h.u21",
                "1prog_l.u22"
            ]
            return dict(zip(filenames, chunks))
        func_map['maincpu'] = maincpu

        # gfx1_1
        def gfx1_1(in_files): 
            contents = in_files['CGCMN.IMG']
            chunks = []
            filenames = [
                # Identical
                "1l.u4",
                "1h.u12",

                # Appears to have matches, but junk thrown in?
                "2l.u5",
                "2h.u13",
            ]

            # 1l.u4 / 1h.u12 - easy
            temp_contents = transforms.cut(contents, 0, 0x400000)
            temp_chunks = transforms.deinterleave(temp_contents, num_ways=2, word_size=2)
            chunks += temp_chunks

            # filenames = [
            #     "odd",
            #     "even"
            # ]
            # chunks = transforms.deinterleave(contents, num_ways=2, word_size=2)
            return dict(zip(filenames, chunks))
        func_map['gfx1_1'] = gfx1_1

        return {'filename': game_info['filename'], 'contents': helpers.build_rom(in_files, func_map)}

        
    def _handle_copyorig(self, in_files, game_info):
        func_map = {}

        # Temporary working
        def raw_files(in_files): 
            return {f'orig/{k}':v for (k,v) in in_files.items()}
        func_map["raw"] = raw_files

        return {'filename': game_info['filename'], 'contents': helpers.build_rom(in_files, func_map)}



'''Implementation of wonderboy3: Wonder Boy The Dragon's Trap'''
from gex.lib.tasks.copytask import CopyTask
from gex.lib.tasks import helpers

class WonderBoyTask(CopyTask):
    '''Implements wonderboy3: Wonder Boy The Dragon's Trap'''
    _task_name = "wonderboy3"
    _title = "Wonder Boy The Dragon's Trap"
    _details_markdown = '''
This is just sitting in the install folder. 

Note that the white box instead of the Sega logo is an intentional change for this release.
    '''
    _default_input_folder = helpers.gen_steam_app_default_folder("Wonder Boy The Dragon's Trap")
    _input_folder_desc = "Wonder Boy The Dragon's Trap Steam Folder"



'''Implementation of mkak: Mortal Kombat Arcade Kollection'''
import logging
import os
from gex.lib.tasks.basetask import BaseTask
from gex.lib.tasks import helpers
from gex.lib.utils.blob import transforms

logger = logging.getLogger('gextoolbox')

class MKAKTask(BaseTask):
    '''Implements mkak: Mortal Kombat Arcade Kollection'''
    _task_name = "mkak"
    _title = "Mortal Kombat Arcade Kollection"
    _details_markdown = '''
    None of these games currently are playable, as it appears the Audio ROMs are replaced with a different audio format.

    It might be possible to convert the Audio to Ogg via instructions at:
    https://borderlands.fandom.com/wiki/Extracting_PC_resources
    Turning those into a functional (let alone CRC matching) ROM is a difficult task.
    See GH issue #25 for the most recent notes.

    '''
    _default_input_folder = helpers.gen_steam_app_default_folder("Mortal Kombat Arcade Kollection")
    _input_folder_desc = "Mortal Kombat Arcade Kollection Steam Folder"

    _out_file_list = [
        {
            "game": "Mortal Kombat",
            "system": "Arcade",
            "filename": "mkla4.zip",
            "status": "partial",
            "notes": [1]
        },
        {
            "game": "Mortal Kombat 2",
            "system": "Arcade",
            "filename": "mk2.zip",
            "status": "partial",
            "notes": [1]
        },
        {
            "game": "Ultimate Mortal Kombat 3",
            "system": "Arcade",
            "filename": "umk3.zip",
            "status": "partial",
            "notes": [1]
        }
    ]
    _out_file_notes = {
        "1": "This ROM is missing the Audio ROM files. MKAK appears to replace it with a different audio solution."
    }
    _game_info_map = {
    }

    _prop_info = {
        "include-partials": {
            "description": "Include the partial ROMs that are missing data; useful for mixing with other sources or investigation",
            "default": False,
            "type": "Boolean"
        }
    }
    def _find_files(self, base_path):
        new_paths = []
        for filename in self._game_info_map:
            new_path = os.path.join(base_path, filename)
            if os.path.exists(new_path):
                new_paths.append(new_path)
            else:
                logging.warning(f"Could not find {filename} in {base_path}")
        return new_paths

    def execute(self, in_dir, out_dir):
        if self._props.get('include-partials'):
            out_files = []
            out_files += self._handle_mk1(in_dir)
            out_files += self._handle_mk2(in_dir)
            out_files += self._handle_umk3(in_dir)

            for out_file in out_files:
                with open(os.path.join(out_dir, out_file['filename']), "wb") as write_file:
                    write_file.write(out_file['contents'])

            logger.info("Processing complete.")
        else:
            logger.error("mkak is not fully implemented - the audio ROMs are missing and have been replaced!")
            logger.error("To help investigate/pull partial contents, add --prop 'include-partials=True' to extract the game and graphics ROMs.")

    def _handle_mk1(self, in_dir):
        file_path = r"BINARIES\WIN32\DATA\MK1"
        in_filenames = [
            "MK1.BIN",
            "MK1AUDIO.BIN",
            "MK1COMP.BIN",
            "MK1IMAGE.BIN"
        ]

        in_files = {}
        for in_filename in in_filenames:
            with open(os.path.join(in_dir, file_path, in_filename), 'rb') as in_file:
                in_files[in_filename] = in_file.read()

        zip_files = {}

        # MAINCPU
        contents = in_files['MK1.BIN']
        chunks = transforms.deinterleave(contents, 2, 1)
        zip_files['l4_mortal_kombat_game_rom_u-105.u105'] = chunks[0]
        zip_files['l4_mortal_kombat_game_rom_u-89.u89'] = chunks[1]

        # GFX
        contents = in_files['MK1IMAGE.BIN']
        chunks = transforms.equal_split(contents, 12)
        gfx_filenames = [
            "l1_mortal_kombat_game_rom_u-111.u111",
            "l1_mortal_kombat_game_rom_u-112.u112",
            "l1_mortal_kombat_game_rom_u-113.u113",
            "l1_mortal_kombat_game_rom_u-114.u114",
            "l1_mortal_kombat_game_rom_u-95.u95",
            "l1_mortal_kombat_game_rom_u-96.u96",
            "l1_mortal_kombat_game_rom_u-97.u97",
            "l1_mortal_kombat_game_rom_u-98.u98",
            "l1_mortal_kombat_game_rom_u-106.u106",
            "l1_mortal_kombat_game_rom_u-107.u107",
            "l1_mortal_kombat_game_rom_u-108.u108",
            "l1_mortal_kombat_game_rom_u-109.u109"
        ]
        zip_files.update(dict(zip(gfx_filenames, chunks)))

        return [{'filename': "partial_mkla4.zip", 'contents': helpers.build_zip(zip_files)}]

    def _handle_mk2(self, in_dir):
        file_path = r"BINARIES\WIN32\DATA\MK2"
        in_filenames = [
            "MK2.BIN",
            "MK2AUDIO.BIN",
            "MK2COMP.BIN",
            "MK2IMAGE.BIN"
        ]

        zip_files = {}
        for in_filename in in_filenames:
            with open(os.path.join(in_dir, file_path, in_filename), 'rb') as in_file:
                zip_files[in_filename] = in_file.read()


        return [{'filename': "partial_mk2.zip", 'contents': helpers.build_zip(zip_files)}]

    def _handle_umk3(self, in_dir):
        file_path = r"BINARIES\WIN32\DATA\UMK3"
        in_filenames = [
            "UMK3.BIN",
            "UMK3AUDIO.BIN",
            "MK4COMP.BIN",
            "UMK3IMAGE.BIN"
        ]

        zip_files = {}
        for in_filename in in_filenames:
            with open(os.path.join(in_dir, file_path, in_filename), 'rb') as in_file:
                zip_files[in_filename] = in_file.read()


        return [{'filename': "partial_umk3.zip", 'contents': helpers.build_zip(zip_files)}]



'''Implementation of neogeo_classics_humble: Neo Geo Classics by SNK Playmore on Humble'''
import logging
import os
import lzma
from gex.lib.tasks.basetask import BaseTask
from gex.lib.tasks import helpers
from gex.lib.utils.blob import transforms

logger = logging.getLogger('gextoolbox')

class NeoGeoClassicsHumbleTask(BaseTask):
    '''Implements neogeo_classics_humble: Neo Geo Classics by SNK Playmore on Humble'''
    _task_name = "neogeo_classics_humble"
    _title = "Neo Geo Classics by SNK Playmore on Humble Store"
    _details_markdown = '''
This task covers a variety of SNK Neo Geo releases Humble Store.
In some cases, these games simply have ZIP files for the ROM; in other cases, the files are in a subfolder.

This also covers most collections/bundles, but only if they install as separate titles. 
As of right now, the only exception known is Samurai Shodown Neogeo Collection.
    '''
    _default_input_folder = r"C:\Program Files (x86)\GOG Galaxy\Games"
    _input_folder_desc = "Game Library (Amazon, Steam, etc.) folder"

    def get_out_file_info(self):
        '''Return a list of output files'''
        return {
            "files": self._metadata['out']['files'],
            "notes": self._metadata['out']['notes']
        }

    def _check_dirs(self, in_dir, child_dirs):
        for child_dir in child_dirs:
            path = os.path.join(in_dir, child_dir)
            if os.path.exists(path):
                return path
        return None

    def _verify_input_zip(self, contents, entries):
        zip_metas = zip_lib.get_metadata(contents)
        # Ensure the file name lists are the same
        real_filenames = set(zip_metas.keys())
        expected_filenames = set(entries.keys())
        if real_filenames != expected_filenames:
            logger.info(f"Could NOT verify {file_name}: File lists don't match!")
            return False
        # Compare file size/CRC
        for inner_filename, zip_meta in zip_metas.items():
            verify_entry = entries[inner_filename]
            if zip_meta['crc'] != verify_entry['crc'] or zip_meta['size'] != verify_entry['size'] :
                logger.info(f"Could NOT verify {file_name}: {inner_filename} should be {verify_entry['crc']} at {verify_entry['size']} bytes, found {zip_meta['crc']} at {zip_meta['size']}")
                return False
        logger.info(f"Verified {file_name}.")
        return True

    def execute(self, in_dir, out_dir):
        in_files = self._metadata['in']['files']

        current_neogeo_bios_ver = None

        # Find folders
        folders = list(set(map(lambda x: x['rel_path'][0], in_files.values())))
        found_folders = []
        for folder in folders:
            path = os.path.join(in_dir, folder)
            if os.path.exists(path):
                logger.info(f"Found {folder}...")
                found_folders.append(folder)
            else:
                logger.info(f"No {folder} folder found...")

        # Process titles
        out_files = {}
        for folder in found_folders:
            title_in_files = [v for v in in_files.values() if v['rel_path'][0] == folder]

            title_neogeo = next(v for v in title_in_files if v['filename'] == "neogeo.zip")
            title_gameroms = [v for v in title_in_files if v['filename'] != "neogeo.zip"]

            for title_gamerom in title_gameroms:
                zip_path = os.path.join(in_dir, *title_gamerom['rel_path'], title_gamerom['filename'])
                if os.path.exists(zip_path):
                    logger.info(f"Found zip {zip_path}...")
                    read_file_entry = self.read_datafile(in_dir, title_gamerom)
                    out_files[title_gamerom['filename']] = read_file_entry['contents']

            if title_neogeo:
                title_bios_version = list(title_neogeo['versions'].keys())[0]
                if title_bios_version == 'enhanced' and current_neogeo_bios_ver != 'enhanced':
                    zip_path = os.path.join(in_dir, *title_neogeo['rel_path'], title_neogeo['filename'])
                    if os.path.exists(zip_path):
                        logger.info(f"Found enhanced NeoGeo BIOS zip {zip_path}...")
                        read_file_entry = self.read_datafile(in_dir, title_neogeo)
                        out_files[title_neogeo['filename']] = read_file_entry['contents']
                        current_neogeo_bios_ver = 'enhanced'
                elif title_bios_version == 'basic' and current_neogeo_bios_ver is None:
                    zip_path = os.path.join(in_dir, *title_neogeo['rel_path'], title_neogeo['filename'])
                    if os.path.exists(zip_path):
                        logger.info(f"Found basic NeoGeo BIOS zip {zip_path}...")
                        read_file_entry = self.read_datafile(in_dir, title_neogeo)
                        out_files[title_neogeo['filename']] = read_file_entry['contents']
                        current_neogeo_bios_ver = 'basic'


        # Write out the files
        for filename, contents in out_files.items():
            _ = self.verify_out_file(filename, contents)
            out_path = os.path.join(out_dir, filename)
            with open(out_path, "wb") as out_file:
                logger.info(f"Writing verified {filename}...")
                out_file.write(contents)


# BSTARS2
# Dotemu2mame.js has this title
# https://gist.github.com/cxx/81b9f45eb5b3cb87b4f3783ccdf8894f


# Looks like Humble release removes team locations

# ZOOM TABLE
# bstars2_zoom_table -> 000-lo.lo

# NEOGEO BIOS
# bstars2_bios_m68k -> neo-epo.bin

# NEOGEO SFIX
# bstars2_bios_sfix -> sfix.sfix
#     transform: NeoGeo.sfix_reorder


# function sfix_reorder(bin)
# {
#     const tmp = Buffer.allocUnsafe(32);
#     for (let i = 0; i < bin.length; i += 32) {
#         for (let j = 0; j < 8; j++) {
#             tmp[j+16] = bin[i+4*j+0];
#             tmp[j+24] = bin[i+4*j+1];
#             tmp[j+ 0] = bin[i+4*j+2];
#             tmp[j+ 8] = bin[i+4*j+3];
#         }
#         tmp.copy(bin, i);
#     }
#     return bin;
# }

# YAMAHA SOUND
# bstars2_adpcm

#     const ymsnd_files = [];
#     const ymsnd_size = fs.statSync(path.join(srcdir, `${name}_adpcm`)).size;
#     for (let i = 1; i <= Math.ceil(ymsnd_size/conf.rom_size[1]); i++)
#         ymsnd_files.push(`${id}-v${i}.v${i}`);

#         ymsnd: {
#             input: `${name}_adpcm`, output: ymsnd_files,
#             transform: bin => split(bin, conf.rom_size[1])
#         },

# SPRITES
# bstars2_tiles

#     const sprites_size = fs.statSync(path.join(srcdir, `${name}_tiles`)).size;
#     const sprites_files = [];
#     for (let i = 1; i <= Math.ceil(sprites_size/2/conf.rom_size[2])*2; i++)
#         sprites_files.push(`${id}-c${i}.c${i}`);


#         sprites: {
#             input: `${name}_tiles`, output: sprites_files,
#             transform: bin => {
#                 NeoGeo.deoptimize_sprites(bin);
#                 if ('cmc_encrypt_gfx' in conf)
#                     conf.cmc_encrypt_gfx(bin);
#                 const a = interleave(bin).map(b => split(b, conf.rom_size[2]));
#                 return a[0].flatMap((b, i) => [b, a[1][i]]);
#             }
#         }


# // from http://i486.mods.jp/ichild/?page_id=62
# // Original author: Imaha486
# function deoptimize_sprites(buf)
# {
#     const tmp = Buffer.allocUnsafe(0x80);
#     for (let i = 0; i < buf.length; i += 0x80) {
#         tmp.fill(0);
#         for (let y = 0; y < 0x10; y++) {
#             let dstData;
#             dstData = buf[i+(y*8)+0] <<  0 |
#                       buf[i+(y*8)+1] <<  8 |
#                       buf[i+(y*8)+2] << 16 |
#                       buf[i+(y*8)+3] << 24;
#             for (let x = 0; x < 8; x++) {
#                 tmp[0x43 | y << 2] |= (dstData >> x*4+3 & 1) << 7-x;
#                 tmp[0x41 | y << 2] |= (dstData >> x*4+2 & 1) << 7-x;
#                 tmp[0x42 | y << 2] |= (dstData >> x*4+1 & 1) << 7-x;
#                 tmp[0x40 | y << 2] |= (dstData >> x*4+0 & 1) << 7-x;
#             }

#             dstData = buf[i+(y*8)+4] <<  0 |
#                       buf[i+(y*8)+5] <<  8 |
#                       buf[i+(y*8)+6] << 16 |
#                       buf[i+(y*8)+7] << 24;
#             for (let x = 0; x < 8; x++) {
#                 tmp[0x03 | y << 2] |= (dstData >> x*4+3 & 1) << 7-x;
#                 tmp[0x01 | y << 2] |= (dstData >> x*4+2 & 1) << 7-x;
#                 tmp[0x02 | y << 2] |= (dstData >> x*4+1 & 1) << 7-x;
#                 tmp[0x00 | y << 2] |= (dstData >> x*4+0 & 1) << 7-x;
#             }
#         }
#         tmp.copy(buf, i);
#     }
#     return buf;
# }


# MAINCPU
# bstars2_game_m68k

#     rom_size: [ 0x080000, 0x100000, 0x100000 ]

#     const maincpu_files = [`${id}-p1.p1`];
#     if (maincpu.length > conf.rom_size[0]) {
#         if (maincpu.slice(conf.rom_size[0]).every(b => b == 0))
#             maincpu = maincpu.slice(0, conf.rom_size[0]);
#         else
#             maincpu_files.push(`${id}-p2.sp2`);
#     }
#         maincpu: {
#             input: maincpu, output: maincpu_files,
#             transform: bin => {
#                 if (conf.swap_68k) {
#                     const a = split_at(bin, bin.length/2);
#                     bin = Buffer.concat([a[1], a[0]]);
#                 }
#                 if ('sma_encrypt' in conf)
#                     conf.sma_encrypt(bin);

#                 const rom_size = conf.rom_size[0];
#                 if (bin.length == rom_size) return bin;

#                 const p1_size = bin.length >= 0x100000 + rom_size ?
#                       0x100000 : bin.length - rom_size;

#                 const a = split_at(bin, p1_size);
#                 if ('sma_encrypt' in conf)
#                     a[0] = a[0].slice(0x0c0000);
#                 return [a[0], ...split(a[1], rom_size)];
#             }
#         },

# GAME SFIX
# bstars2_game_sfix -> 041-s1.s1
#     transform: NeoGeo.sfix_reorder


# function sfix_reorder(bin)
# {
#     const tmp = Buffer.allocUnsafe(32);
#     for (let i = 0; i < bin.length; i += 32) {
#         for (let j = 0; j < 8; j++) {
#             tmp[j+16] = bin[i+4*j+0];
#             tmp[j+24] = bin[i+4*j+1];
#             tmp[j+ 0] = bin[i+4*j+2];
#             tmp[j+ 8] = bin[i+4*j+3];
#         }
#         tmp.copy(bin, i);
#     }
#     return bin;
# }

# AUDIOCPU
# bstars_game_m68k -> 041-m1.m1


        logger.info("Processing complete.")



import click
import pkg_resources

@click.command()
def version():
    """Return build info"""
    version = pkg_resources.get_distribution('game-extraction-toolbox').version
    print(f'Game Extraction Toolbox v{version}')



'''Implementation of basic copy-only task'''
import logging
import os
from gex.lib.tasks.basetask import BaseTask

logger = logging.getLogger('gextoolbox')

class CopyTask(BaseTask):
    '''Implements basic copy-only task'''

    def get_out_file_info(self):
        '''Return a list of output files'''
        return {
            "files": self._metadata['out']['files'],
            "notes": self._metadata['out']['notes']
        }

    def execute(self, in_dir, out_dir):
        '''Copy/rename the ROM files'''
        for file_metadata in self._metadata['in']['files'].values():
            resolved_file = self.read_datafile(in_dir, file_metadata)
            out_file_entry = [x for x in self._metadata['out']['files'] if x['game'] == file_metadata['copy_to']][0]

            filename = out_file_entry['filename']
            _ = self.verify_out_file(filename, resolved_file['contents'])
            out_path = os.path.join(out_dir, filename)
            with open(out_path, "wb") as out_file:
                out_file.write(resolved_file['contents'])
        logger.info("Processing complete.")


'''Template task to implment tasks with'''
import logging

from gex.lib.tasks.basetask import BaseTask

logger = logging.getLogger('gextoolbox')

class MyTask(BaseTask):
    '''Implments xxxxxx: Xxxx Xxxx Xxxx'''
    _task_name = ""
    _title = ""
    _details_markdown = '''

    '''
    _default_input_folder = r""
    _input_folder_desc = ""


    def execute(self, in_dir, out_dir):
        '''Main implementation call for the extraction task'''
        logger.info("Processing complete.")



import json
import logging
import os
import click
import click_log
from texttable import Texttable
from gex.lib.utils.blob import hash as hash_helper

logger = logging.getLogger('gextoolbox')

@click.command()
@click.option('--1', 'file_1', help = 'path to input file 1', required=True)
@click.option('--2', 'file_2', help = 'path to input file 2', required=True)
@click_log.simple_verbosity_option(logger)
def diff(file_1, file_2):
    '''Smarter non-interactive diff tool'''
    try:
        with open(file_1, "rb") as file:
            contents_1 = file.read()
    except IOError:
        logger.error(f"Error reading {file_1}!")
        exit()
    try:
        with open(file_2, "rb") as file:
            contents_2 = file.read()
    except IOError:
        logger.error(f"Error reading {file_2}!")
        exit()


    table = Texttable()
    table.add_row(["", "File 1", "File 2", ""])
    table.set_cols_dtype(["t", "t", "t", "t"])
    filename_1 = os.path.basename(file_1)
    filename_2 = os.path.basename(file_2)
    table.add_row(['filename', filename_1, filename_2, filename_1 == filename_2])
    dir_1 = os.path.abspath(os.path.dirname(file_1))
    dir_2 = os.path.abspath(os.path.dirname(file_2))
    table.add_row(['path', dir_1, dir_2, dir_1 == dir_2])

    # Diff size
    size_1 = len(contents_1)
    size_2 = len(contents_2)
    if size_1 == size_2:
        logger.info(f"File sizes are identical at {size_1} bytes...")
    else:
        logger.info(f"File sizes are different, {size_1} vs {size_2} bytes...")
    table.add_row(['size', size_1, size_2, size_1 == size_2])

    # Diff CRC
    crc_1 = hash_helper.get_crc(contents_1)
    crc_2 = hash_helper.get_crc(contents_2)
    if crc_1 == crc_2:
        logger.info(f"File hashes are identical, {crc_1}...")
    else:
        logger.info(f"File hashes are different, {crc_1} vs {crc_2}...")
    table.add_row(['crc', crc_1, crc_2, crc_1 == crc_2])

    # TODO: Find sliding match if diff size

    # Simple percent comparison
    if size_1 == size_2:
        diff_bytes = {}
        for x in range(0, size_1):
            if contents_1[x] != contents_2[x]:
                diff_bytes[x] = {"1": contents_1[x], "2": contents_2[x]}

        identical_percent = (100 - (len(diff_bytes.keys()) / size_1)) / 100
        
        logger.info(f"Percent similar: {identical_percent:.4%}")

        # Make a patch-y file
        # TODO: Make this a flag
        patch_data = {int(k): v["1"] for k, v in diff_bytes.items()}
        print(json.dumps(patch_data))
    else:
        logger.info("Percentage comparison for differently sized files NYI!")

    print(table.draw())

